#ifndef __DATA_H_INCLUDED__
#define __DATA_H_INCLUDED__

#include <cstring>

class ClassInfoDataDesc;

// a data descriptor
class DataDesc {
public:
	virtual ~DataDesc(void) {}

	// GetSample provides a reference to the sample specified by i. 
	// The sample itself is stored in memory space that is managed by the DataDesc. 
	// Calling Flush() invalidates the reference provided by GetSample.
	virtual void GetSample  (double** p, int i)                   = 0;
	
	// Does exactly what GetSample does but for a set of len data samples that 
	// are specified by I. 
	virtual void GetSamples (double** P, int* I, int len) = 0;

	// CopySample copies the valid dimensions of the sample specified by i to *x.
	// NZ is set either to NULL or to some array which contains the valid dimensions. 
	virtual int CopySample (double* x, int** NZ, int i)           = 0;

	virtual void PutSample(double* x, int* NZ, int NZ_len, int i) = 0;

	// returns the number of dimensions of each sample.
	virtual int ssize (void)       = 0;
	// returns the number of samples
	virtual int num_samples (void) = 0;
	int size(void) { return num_samples(); }

	// returns the number of values stored in the data descriptor.
	virtual int Count(void)        = 0;

	// Invalidates (frees) all temporary memory that is managed by the data descriptor.
	virtual void Flush(void)       = 0;

	virtual void getAbsIndex (int* index) = 0;
	virtual void getAbsIndex (int* index, int* I, int len) = 0;

	virtual DataDesc* ID(void) = 0;

	virtual void GetAsMatrix(double*** ret);
};

// A Proxy.
class ProxyDataDesc : public DataDesc {
public:
	ProxyDataDesc(DataDesc* D) : pdesc(D) {}

	virtual ~ProxyDataDesc(void) {}
	
	virtual void GetSample  (double** p, int i) { pdesc->GetSample  (p, i); }
	virtual void GetSamples (double** P, int* I, int len) { pdesc->GetSamples (P, I, len); }

	virtual int CopySample (double* x, int** NZ, int i) { return pdesc->CopySample (x, NZ, i); }

	virtual void PutSample(double* x, int* NZ, int NZ_len, int i) { pdesc->PutSample(x, NZ, NZ_len, i); }

	virtual int ssize (void) { return pdesc->ssize(); }
	virtual int num_samples (void) { return pdesc->num_samples(); }

	virtual int Count(void) { return pdesc->Count(); }
	virtual void Flush(void) { return pdesc->Flush(); }

	virtual void getAbsIndex (int* index) { pdesc->getAbsIndex(index); };
	virtual void getAbsIndex (int* index, int* I, int len) { pdesc->getAbsIndex(index, I, len); };

	virtual DataDesc* ID(void) { return pdesc->ID(); };
protected:
	DataDesc* pdesc;
};

// A wrapper that removes the vector m from each sample contained in the wrapped data set
class MeanRemover : public ProxyDataDesc {
public:
	MeanRemover(double* m, DataDesc* D, int save_memory);

	virtual ~MeanRemover(void);

	virtual void GetSample  (double** p, int i);

	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);

	virtual void Flush(void);

private:
	DataDesc* data;

	int* removed;
	int* I_removed;
	int num_removed;

	int save_memory;
	double* mean;
};

// MemStore manages the temporary memory that is used by several types of data descriptors
class MemStore {
public:
	MemStore(int size, int esize);
	virtual ~MemStore (void);

	virtual double* ref(int i);
	virtual int is_in_mem(int i);
	virtual void flush(void);

	virtual void set_entry_size(int new_size);

private:
	double** in_mem;
	int* i_in_mem;
	int in_mem_size;

	int size;
	int esize;
};

// A data descriptor wrapper of a simple double** matrix.
class StandardDataDesc : public DataDesc {
public:
	StandardDataDesc (double** X, int esize, int nentrys, int isOwner=1);
	~StandardDataDesc(void);

	virtual int ssize(void);
	virtual int num_samples (void);	
	virtual int Count(void);

	virtual void GetSample(double** x, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual void Flush(void); // empty function (does not invalidate anything)

	virtual void getAbsIndex (int* index);
	virtual void getAbsIndex (int* index, int* I, int len);

	virtual DataDesc* ID(void) { return this; }

private:
	double** D;
	int size;
	int entry_size;

	int* test;
	int isOwner;
};


class ImageDataDesc : public DataDesc {
public:
	ImageDataDesc(double** image, int nentrys, int esize, int tile_nentrys, int tile_esize, int random_tile_count =0, int isOwner=1);
	~ImageDataDesc(void);

	virtual int ssize(void);
	virtual int num_samples (void);	
	virtual int Count(void);

	virtual void GetSample(double** x, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual void Flush(void); // empty function (does not invalidate anything)

	virtual void getAbsIndex (int* index);
	virtual void getAbsIndex (int* index, int* I, int len);

	virtual DataDesc* ID(void) { return this; }
private:
	double** image;
	int nentrys;
	int esize;

	int tile_nentrys;
	int tile_esize;

	int random_tile_count;
	int isOwner;
};

// A data descriptor that implements sparse matrices.
class SparseDataDesc : public DataDesc {
public:
	SparseDataDesc (char* directory);
	SparseDataDesc (int nentrys, int entry_size);

	virtual ~SparseDataDesc(void);

	virtual int ssize(void);
	virtual int num_samples (void);	
	virtual int Count(void);
	
	virtual void GetSample(double** x, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample(double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual void Flush(void);

	virtual void getAbsIndex (int* index);
	virtual void getAbsIndex (int* index, int* I, int len);

	virtual DataDesc* ID (void) { return this; }

	//---------

	virtual SparseDataDesc* transpose(void);
	virtual SparseDataDesc* copy(void);

	virtual void put(SparseDataDesc* src);
	virtual void get(SparseDataDesc* dest); 

	virtual void mean (double* data_mean);
	virtual void sample_mean(double* smean);
	virtual void get_norm (double* norm);

	virtual double mean_var(void);

	virtual void smultiply (double* fac);
	virtual void smultiply(double s);

	virtual void ssub(double* sum);
	virtual void sub(double* sum);

	virtual void delete_entrys(double value);
	virtual double count_entrys(double value);

	virtual void save(char* directory);
	virtual void get_matrix(double*** M, int* nentrys, int* esize);

protected:
	SparseDataDesc (int nsamples, int entry_size, double** D_tmp, int** active_dims_tmp, int* active_dim_len_tmp);

	double** D;
	int** active_dims;
	int* active_dim_len;	
private:
	int nsamples;
	int entry_size;

	MemStore* store;
};

// A wrapper class for any other data descriptor that adds class information
// to the data.
class ClassInfoDataDesc : public ProxyDataDesc {
protected:
	ClassInfoDataDesc(DataDesc* desc);
public:
	ClassInfoDataDesc(double* ct, int ctlen, DataDesc* desc);
	ClassInfoDataDesc(char* data_file, char* class_info_file, int save_memory=0);

	virtual ~ClassInfoDataDesc(void);

	virtual void getClassInfo(double* c_dest);
	virtual double getClassInfo(int i);
	virtual void getClassInfo(double* c_dest, int* I, int I_len);
protected:
	double* c;
};


// This data descriptor selects a subset of some given data desciptor
class SubsetDataDesc : public ProxyDataDesc {
public:
	SubsetDataDesc (int* S_tmp, int S_len, DataDesc* D);

	virtual ~SubsetDataDesc(void);
	
	virtual void GetSample  (double** p, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);

	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual int num_samples (void);

	virtual void getAbsIndex (int* index);
	virtual void getAbsIndex (int* index, int* I, int len);

private:
	int* S;
	int S_len;
};

// A data descriptor that selects a subset of dimensions of given data
class DimSelectDataDesc : public ProxyDataDesc {
public:
	DimSelectDataDesc(int* M_tmp, int M_len, DataDesc* desc, int save_memory=0);	

	virtual ~DimSelectDataDesc (void);

	virtual void GetSample  (double** p, int i);			// GetSample should be preferred because it does not touch the entire sample (it only provides a reference)
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);

	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual int ssize(void);

	virtual void Flush(void);

	virtual void remove_dims (int* d, int count);
	virtual void GetDimIndex (int* dest, int* dims, int len);	
	virtual void GetIndex (int* dest);

private:
	int* M;
	int M_len;

	int save_memory;

	MemStore* store;
};


class RemoveNANsDataDesc : public ProxyDataDesc {
public:
	RemoveNANsDataDesc(DataDesc* desc);	

	virtual ~RemoveNANsDataDesc (void);

	virtual int CopySample (double* x, int** NZ, int i);

private:
	int* valid_dims;
};



// A data descriptor that selects a subset of dimensions of given data
class ClassInfoDimSelectDataDesc : public ClassInfoDataDesc {
public:
	ClassInfoDimSelectDataDesc(int* M_tmp, int M_len, ClassInfoDataDesc* desc);	

	void remove_dims (int *d, int count);
	void GetDimIndex (int* dest, int* dims, int len);	
	void GetIndex (int* dest);

	DimSelectDataDesc* GetDimSelector(void) { return dim_select; }
private:
	DimSelectDataDesc* dim_select;
};

// This data describtor selects a subset of some given data desciptor that has class information
class SubsetClassInfoDataDesc : public ClassInfoDataDesc {
public:
	SubsetClassInfoDataDesc(int* S_tmp, int S_len, ClassInfoDataDesc* D);
	virtual ~SubsetClassInfoDataDesc(void);

	virtual void getClassInfo(double* c_dest);
	virtual double getClassInfo(int i);
	virtual void getClassInfo(double* c_dest, int* I, int I_len);
};

// A data descriptor that loads samples subsequently from a file.
class DiskDataDesc : public DataDesc {
public:
	DiskDataDesc(char* fname, int save_memory=0);

	virtual ~DiskDataDesc (void);

	virtual void GetSample  (double** p, int i);
	virtual void GetSamples (double** P, int* I, int len);

	virtual int CopySample (double* x, int** NZ, int i);
	virtual void PutSample(double* x, int* NZ, int NZ_len, int i);

	virtual int ssize (void);
	virtual int num_samples (void);
	virtual int Count(void);

	virtual void Flush(void);

	virtual void getAbsIndex (int* index);
	virtual void getAbsIndex (int* index, int* I, int len);

	virtual DataDesc* ID(void) { return this; }

private:
	char* filename;
	double* file_index;
	int file_index_len;

	int entry_size;

	MemStore* store;
	int save_memory;
};

#endif
