#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <regex.h>
#include <cmath>
#include <errno.h>

#include "utils.h"
#include "Data.h"
#include "MXData.h"
#include "DiskUtils.h"

ClassInfoDataDesc* MXDataDesc::Create(const mxArray* mx_data, const mxArray* mx_c, int save_memory) {

	DataDesc* data = Create(mx_data, save_memory);

	int c_size = mxGetM(mx_c)*mxGetN(mx_c);

	ClassInfoDataDesc* ret = new ClassInfoDataDesc (mxGetPr(mx_c), c_size, data);

	return ret;
}

DataDesc* MXDataDesc::Create(const mxArray* mx_data, int save_memory) {

	DataDesc* ret = NULL;

	mxClassID id_data = mxGetClassID (mx_data);

	if (id_data == mxCHAR_CLASS) {

		int buflen = mxGetM(mx_data)*mxGetN(mx_data)*sizeof(mxChar) + 1;
		char* filename   = char_array(buflen);

		mxGetString(mx_data, filename, buflen);

		ret = new DiskDataDesc(filename, save_memory);

		free_array(filename);
	}
	else if(id_data == mxDOUBLE_CLASS) {
		int X_rows    = mxGetM (mx_data);
		int X_columns = mxGetN (mx_data);

		double** pX = (double**) pointer_array (X_columns);
		double* pX_base = mxGetPr(mx_data);
		for (int i=0; i < X_columns; i++) pX[i] = pX_base + i*X_rows;

		double** X = m_create(X_columns, X_rows);
		m_copy (X, pX, X_columns, X_rows);

		free_array(pX);

		ret = new StandardDataDesc (X, X_rows, X_columns);
	}
	else if (id_data == mxUINT8_CLASS) {
		ret = new MXUINT8DataDesc(mx_data, save_memory);
	}
	else if (id_data == mxINT32_CLASS) {
		ret = new MXINT32DataDesc(mx_data, save_memory);
	}
	else PRINTF("ERROR!!!! Data type not supported!\n");

	return ret;
}

SubsetDataDesc* MXDataDesc::Create(DataDesc* desc, const mxArray* mx_train_index) {

	int* train_index  = NULL;
	int tr_index_size = (mx_train_index != NULL) ? mxGetN(mx_train_index)*mxGetM(mx_train_index) : 0;

	if (tr_index_size) { // consider subset
		train_index = int_array(tr_index_size);

		if (mxGetClassID (mx_train_index) == mxINT32_CLASS) memcpy(train_index, (int*) mxGetData(mx_train_index), tr_index_size*4); 
		else double2int (train_index, mxGetPr(mx_train_index), tr_index_size);

		issub_ip(train_index, 1, tr_index_size); // need zero based indices

		SubsetDataDesc* ret = new SubsetDataDesc (train_index, tr_index_size, desc);

		free_array(train_index);

		return ret;
	}
	else return NULL;
}

/////////////////////////////////////
//////
/////////////////////////////////////

MXUINT8DataDesc::MXUINT8DataDesc(const mxArray* mx_data, int save_memory) : MXDataDesc(mx_data), D(NULL), store(NULL), save_memory(save_memory) {

	D = (unsigned char**) pointer_array (num_samples());

	D[0] = (unsigned char*) mxGetData(mx_data);

	for (int i=1; i < num_samples(); i++) D[i] = D[0] + i*entry_size;

	store = new MemStore(size, entry_size);
}

MXUINT8DataDesc::~MXUINT8DataDesc(void) {

	free_array(D);

	delete store;
}

void MXUINT8DataDesc::GetSample(double** p, int i) {

	if (store->is_in_mem(i)) *p = store->ref(i);
	else {
		*p = store->ref(i);

		uchar2double  (*p, D[i], entry_size);
	}

}

void MXUINT8DataDesc::GetSamples (double** P, int* I, int len) {
	for (int i=0; i < len; i++) GetSample(P+i, I[i]);
}

int MXUINT8DataDesc::CopySample (double* x, int** NZ, int i) {
	if (NZ != NULL) *NZ = NULL;

	uchar2double  (x, D[i], entry_size);

	return entry_size;
}

void MXUINT8DataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	if (NZ !=  NULL) {
		double* tmp  = double_array(NZ_len);
		double* tmp2 = double_array(entry_size);

		get_i(tmp, x, NZ, NZ_len);
		uchar2double(tmp2, D[i], entry_size);

		set_i(tmp2, tmp, NZ, NZ_len); 
		double2uchar(D[i], tmp2, entry_size);

		free_array(tmp);
		free_array(tmp2);
	}
	else double2uchar(D[i], x, entry_size); 
}

void MXUINT8DataDesc::Flush(void) {
	if(save_memory) store->flush();
}

/////////////////////////////////////
//////
/////////////////////////////////////

MXINT32DataDesc::MXINT32DataDesc(const mxArray* mx_data, int save_memory) : MXDataDesc(mx_data), D(NULL), store(NULL), save_memory(save_memory) {

	D = (int**) pointer_array (num_samples());

	D[0] = (int*) mxGetData(mx_data);

	for (int i=1; i < num_samples(); i++) D[i] = D[0] + i*entry_size;

	store = new MemStore(size, entry_size);
}

MXINT32DataDesc::~MXINT32DataDesc(void) {

	free_array(D);

	delete store;
}

void MXINT32DataDesc::GetSample(double** p, int i) {

	if (store->is_in_mem(i)) *p = store->ref(i);
	else {
		*p = store->ref(i);

		int2double  (*p, D[i], entry_size);
	}

}

void MXINT32DataDesc::GetSamples (double** P, int* I, int len) {
	for (int i=0; i < len; i++) GetSample(P+i, I[i]);
}

int MXINT32DataDesc::CopySample (double* x, int** NZ, int i) {
	if (NZ != NULL) *NZ = NULL;

	int2double(x, D[i], entry_size);

	return entry_size;
}

void MXINT32DataDesc::PutSample(double* x, int* NZ, int NZ_len, int i) {
	if (NZ !=  NULL) {
		double* tmp  = double_array(NZ_len);
		double* tmp2 = double_array(entry_size);

		get_i(tmp, x, NZ, NZ_len);
		int2double(tmp2, D[i], entry_size);

		set_i(tmp2, tmp, NZ, NZ_len); 
		double2int(D[i], tmp2, entry_size);

		free_array(tmp);
		free_array(tmp2);
	}
	else double2int(D[i], x, entry_size); 
}

void MXINT32DataDesc::Flush(void) {
	if(save_memory) store->flush();
}
