/*
KNN classifier written by Kai Labusch
Copyright (C) 2006 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/


#include <cmath>
#include <cstdlib>
#include <cstring>

#include "utils.h"
#include "min.h"
#include "knn.h"

using namespace std;

/*
X_test is not 16byte aligned !!!
*/
void knn (double*** r_result,  double*** r_I, double** X_train, double* c, int num_train,
	  double** X_test, int num_test, int ssize, int* k, int k_length) {

	int i=0;
	int j=0;
	int l=0;

	/* sort k*/
	qsort((void*) k, k_length, sizeof(int), &icompare);

	int* c_index = int_array(num_train);

	/* determine number of classes */
	double* un = double_array(num_train);
	int num_cl = unique(un, c, num_train);
	int* ccounter = int_array(num_cl);

	/* create map for class labels */
	for (i=0; i < num_train; i++)
		for(j=0; j < num_cl; j++) if(c[i] == un[j]) c_index[i] = j;

	double** result = NULL;
	if (*r_result != NULL) result = *r_result;
	else {
		/* allocate matrix to store result */
		result = m_create (k_length, num_test);
		*r_result = result;	
	}
	m_zero (result, k_length, num_test);

	int k_max      = k[imax(k, k_length)];

	if(r_I != NULL) 
		if (*r_I == NULL) *r_I = m_create(num_test, k_max);

	double* norm = double_array(num_train);
	int* index   = int_array(num_train);

	/* calculate norm of each sample */
	for (i=0;i < num_train; i++) {
		norm[i]  = sqrt(scp(X_train[i],X_train[i], ssize));
		index[i]= i;
	}

	/* sort norm of each training sample */
	isort(norm, index, num_train);

	/* perform classification */
	int left  = 0;
	int right = 0;
	int stop  = 0;

	double* heap   = double_array(k_max);
	int*    heap_I = int_array(k_max);
	int heap_size  = 0;
	double cur_norm = 0;

	double* x_test = double_array(ssize);
	for (i=0;i < num_test;i++) { /* for each sample */
		memcpy(x_test, X_test[i], sizeof(double)*ssize);

		memset(ccounter, 0, num_cl*sizeof(int)); /* reset counters */
		memset(heap    , 0, k_max*sizeof(double)); /* reset heap */
		memset(heap_I  , 0, k_max*sizeof(int));

		heap_size = 0;	
		left = right = -1;
		stop  = 0;
		cur_norm = sqrt(scp(x_test,x_test,ssize));

		while (!stop) stop = insert_knn_sample (X_train, x_test, ssize, heap, heap_I, &heap_size, k_max, num_train,
		      					norm, index, cur_norm, &left, &right);

		if (k_length > 1)  isort(heap, heap_I, heap_size);

		if (r_I != NULL) int2double((*r_I)[i], heap_I, heap_size);	

		l=0;
		for (j=0;j < k_length; j++) { /* for each k-value to test */

			/* consider up to k[j] nearest neighbours */
			while ((l < k[j]) && (l < heap_size)) ccounter[c_index[heap_I[l++]]]++;

			/* find maximal counter entry -> assign class information */
			result[j][i] = un[imax(ccounter, num_cl)];		
		}
	}

	free_array(x_test);	
	free_array(norm);
	free_array(index);
	free_array(un);
	free_array(c_index);
	free_array(ccounter);

	free_array(heap);
	free_array(heap_I);
}

int insert_knn_sample (double** X, double* x_test, int ssize, 
		      double* heap, int* heap_I, int* heap_size, int k, int num_train,
		      double* norm, int* index, double norm_x_test, 
		      int* left, int* right) {

	int new_sample = 0;
	double norm_dist = 0;
	double norm_dist_left = 0;
	double norm_dist_right = 0;
	double dist = 0;

	if ((*left == -1) && (*right == -1)) { /* initialization */

		bin_search (norm_x_test, norm, left, right, num_train);
	
		if (norm_x_test - norm[*left] < norm[*right] - norm_x_test) new_sample = *right = *left;
		else  new_sample = *left = *right;	
		
		dist = sqrt(diff_scp(x_test, X[index[new_sample]], ssize));
		*heap_size = heap_insert (heap, heap_I, dist, index[new_sample], *heap_size);

		(*left)--;
		(*right)++;

		return 0;
	}

	if ((*left < 0) && (*right > num_train - 1)) return 1; /* nothing to do, stop */
	else if ((*left < 0) && (*right <= num_train - 1)) {
		new_sample = (*right)++; 
		norm_dist = norm[new_sample]  - norm_x_test; 
	}
	else if ((*left >= 0) && (*right > num_train - 1)) {
		new_sample = (*left)--;
		norm_dist = norm_x_test - norm[new_sample];
	}
	else {
		norm_dist_left  = norm_x_test - norm[*left];
		norm_dist_right = norm[*right] - norm_x_test;
		if (norm_dist_left < norm_dist_right) {
			new_sample = (*left)--;
			norm_dist = norm_dist_left;
		}
		else {
			new_sample = (*right)++;
			norm_dist = norm_dist_right;
		}
	}

	dist = sqrt(diff_scp(x_test, X[index[new_sample]], ssize));

	if (*heap_size < k) {
		*heap_size = heap_insert (heap, heap_I, dist, index[new_sample], *heap_size);
		return 0;
	}
	else {
		if (heap[0] < norm_dist) return 1; /* abort criterion |a - b| < |a| - |b| */ 

		if (heap[0] > dist) {
			*heap_size = heap_extract_max (heap, heap_I, *heap_size);
			*heap_size = heap_insert (heap, heap_I, dist, index[new_sample], *heap_size);
		}	
		return 0; 
	}
}
