/*
sparse coding neural gas algorithm written by Kai Labusch
Copyright (C) 2009 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/


#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

using namespace std;

#include "utils.h"
#include "ngas.h"
#include "Data.h"

void omp_rec (int k, double* p, double* r, int* next_win, double noise_level, double** codebook, int ssize, int codebook_size, int* candidates, 
	     double* y, double* temp,double* temp2, double* winner, int* next_win_len, int cand_len);

void evaluate_current_solution(double* r_mXE, double* r_mN0, double* r_mAE,
			       int num_components, double noise_level, DataDesc* data,
          		       double** codebook, int codebook_size, double** A_orig);

void momp_rec (double** P, double** R, int k, double* cur_r, double noise_level, 
              double** cur_codebook, int ssize, int codebook_size, int* cur_candidates, int cand_len,
              int* schedule, int* n_SOLUTIONS, int** NEXT_WIN,  int* DEPTH, int depth); 

double** spngas(DataDesc* data, int codebook_size, int epochs, int num_components,
		double alpha0, double lambda0, double alphaf, double lambdaf, double noise_level, int t_start, int t_max, 
		double* E, double* N0, double* AE, double** A_orig, double*** track_codebook,
		int num_track, int operation_mode, double** initial_codebook) {

	int i=0;
	int j=0;
	int k=0;

	int X_columns = data->num_samples();
	int ssize     = data->ssize();

	if (num_components > codebook_size) {
		PRINTF("num_components > codebook_size!!! setting num_components=%g !!!", codebook_size);
		num_components = codebook_size;
	}
	
	double** codebook     = m_create(codebook_size, ssize);
	double** codebook_tmp = m_create(codebook_size, ssize);
	if (initial_codebook == NULL) {
		for(i=0; i < codebook_size; i++) drand_array(codebook[i], -0.5*10e-5, 0.5*10e-5, ssize);
	}
	else m_copy(codebook, initial_codebook, codebook_size, ssize);
	// set norm of each codebook vector to 1 
	for (j=0; j < codebook_size; j++) {
		double n = sqrt(scp(codebook[j], codebook[j], ssize)); 
		if (n==0) n = 1;
		sdivide(codebook[j], codebook[j], n, ssize);
	}

	int train_steps = epochs*X_columns;
	if (t_max == 0) t_max = train_steps;

	int track_steps = 0;
	if (operation_mode == OPERATION_MODE_BATCH) {
		if ((E != NULL) && (N0 != NULL) && (AE != NULL)) {
			track_steps = train_steps/num_track;
			memset(E , 0, num_track*sizeof(double));
			memset(N0, 0, num_track*sizeof(double));
			memset(AE, 0, num_track*sizeof(double));
		}
	}
	else if (num_track > 0) track_steps = train_steps/num_track;

	if (alpha0  < 0) alpha0  = 0.5;
	if (lambda0 < 0) lambda0 = ((double) codebook_size)/2.0;
	
	if (alphaf < 0 ) alphaf  = 0.005;
	if (lambdaf < 0) lambdaf = 0.01;

	double K     = (double) num_components - 1;

	double* temp  = double_array(codebook_size);
	double* y     = double_array(codebook_size);
	int* itemp    = int_array(codebook_size);

	double* ranking = double_array(codebook_size);
	double* h       = double_array(codebook_size);

	double* cdiff = double_array(ssize);	
	double* ctemp = double_array(ssize);	
	double* x     = double_array(ssize);

	int* candidates = int_array(codebook_size);

	double old_percent = 0.0;
	int cb_win         = 0;
	PRINTF("SCNG ...\n");
	PRINTF("number of samples: %i, codebook size: %i, data size: %i, num_components: %i, alpha0: %g, alphaf: %g, lambda0: %g, lambdaf: %g, noise_level: %g\n",
	X_columns, codebook_size, ssize, num_components, alpha0, alphaf, lambda0, lambdaf, noise_level);
	PRINTF("t_start: %i, t_max: %i, number of training samples: %i, numer of tracked time steps: %i\n", t_start, t_max, X_columns, num_track);

	for (i=t_start; i < t_start + train_steps; i++) {
		double percent = floor( ((double)i)*100/train_steps );
		if (old_percent != percent) {
			old_percent = percent;
			PRINTF("%g %% \n", percent);
			FLUSH;
		}

		// compute current learning rate as well as current neighbourhood size
		double lambda = lambdaf;
		double alpha  = alphaf;

		if(i <= t_max) { // for i > t_max use lambdaf, alphaf (online learning)
			lambda = lambda0*pow(lambdaf/lambda0, ((double)i)/((double) t_max));
			alpha  = alpha0*pow(alphaf/alpha0   , ((double)i)/((double) t_max));
		}

		for (j=0; j < codebook_size; j++) { // set norm of each codebook vector to 1
			candidates[j] = j;
			double n      = sqrt(scp(codebook[candidates[j]], codebook[candidates[j]], ssize)); 
			if (n==0)   n = 1;
			sdivide(codebook[candidates[j]], codebook[candidates[j]], n, ssize);
		}

		if (track_steps) {
			if (((i-t_start)/track_steps)*track_steps == i-t_start) {
				int ind = (i-t_start)/track_steps;
				if (operation_mode == OPERATION_MODE_BATCH)  {
					evaluate_current_solution(E+ind, N0+ind, AE+ind, num_components, noise_level, data, codebook, codebook_size, A_orig);
				}
				else {
					track_codebook[ind] = m_create(codebook_size, ssize);
					m_copy(track_codebook[ind], codebook, codebook_size, ssize);	
				}
			}
		}

		// select random sample for learning
		int* active_dims = NULL;
		int selected     = 0;

		if (operation_mode == OPERATION_MODE_BATCH) // in case of batch learning select training samples in random order from entire subset 
			selected = (int) (((X_columns - 1)*(double)rand())/(RAND_MAX + 1.0));
		else	// in case of online learning just process the sample sequence
			selected	 = (i-t_start) % X_columns;

	 	int cur_ssize    = data->CopySample(x, &active_dims, selected);

		if (active_dims == NULL) m_copy (codebook_tmp, codebook, codebook_size, ssize);
		else for (j=0;j < codebook_size;j++) get_i(codebook_tmp[j], codebook[j], active_dims, cur_ssize);

		double nx = sqrt(scp(x,x, cur_ssize)); // compute norm of current residual
		k=0;
		while((k < num_components) && (nx > noise_level) ){ // perform orthogonal matching pursuit
		
			for (j=0; j < codebook_size - k; j++) {				// calculate distance of the residual with respect to all codebook vectors y = w*x 
				y[j]     = scp(x, codebook_tmp[candidates[j]], cur_ssize);
				itemp[j] = j; 
			}
			multiply(temp , y   , y , codebook_size - k); 			// temp = -(y^2)
			smultiply(temp, temp, -1, codebook_size - k); 

			isort(temp, itemp, codebook_size - k);				// calculate neighbourhood ranking
			linspace(temp, 0, codebook_size - 1 - k, codebook_size - k); 	// temp = [0:n-1]
			set_i(ranking, temp, itemp, codebook_size - k);
			smultiply(ranking, ranking, -1.0/lambda, codebook_size - k);
			exponential(h, ranking, codebook_size - k); 

			for (j=0; j < codebook_size - k; j++) {				// calculate update of codebook vectors according to oja's rule
				smultiply(ctemp, codebook_tmp[candidates[j]], y[j], cur_ssize);					// ctemp	   = codebook_tmp[j]*y[j]
				sub(cdiff, x, ctemp, cur_ssize); 			   					// cdiff	   = x - ctemp
				smultiply(cdiff, cdiff, y[j]*alpha*h[j], cur_ssize);   						// cdiff 	   = cdiff*alpha*h[j]*y[j]
				add(codebook_tmp[candidates[j]] , cdiff, codebook_tmp[candidates[j]], cur_ssize); 		// codebook_tmp[j] = codebook_tmp[j]+cdiff

				if (active_dims==NULL) add(codebook[candidates[j]], cdiff, codebook[candidates[j]], ssize); 	// codebook[j]     = codebook[j] + cdiff 
				else {
					set_i(ctemp, cdiff, active_dims, cur_ssize);
					add_ip_i(codebook[candidates[j]], ctemp, active_dims, cur_ssize);	
				}

				// set norm of each codebook vector to 1
				double n = sqrt(scp(codebook_tmp[candidates[j]], codebook_tmp[candidates[j]], cur_ssize)); 
				if (n==0) n = 1;
				sdivide(codebook_tmp[candidates[j]], codebook_tmp[candidates[j]], n, cur_ssize);

				// calculate overlap of current residual with respect to each updated codebook vector  y = w*x
				y[j]     = scp(x, codebook_tmp[candidates[j]], cur_ssize);
				itemp[j] = j;
			}

			multiply (temp, y   , y , codebook_size - k); // temp = -(y^2)
			smultiply(temp, temp, -1, codebook_size - k); 
			
			// use winning codebook vector (largest overlap)
			cb_win       = dmin (temp, codebook_size - k);
			double* pwin = codebook_tmp[candidates[cb_win]];

			double nwin       = sqrt(scp(pwin, pwin, cur_ssize)); 
			if (nwin==0) nwin = 1;

			// project x on winning codebook vector
			smultiply(ctemp, pwin, (y[cb_win]/nwin)/nwin, cur_ssize);

			// subtract the projection from x -> x is now orthogonal with respect to cb_win
			sub(x, x, ctemp, cur_ssize);

			// remove cb_win from the set of candidates
			int rest = iremove_entry(candidates, candidates, cb_win, codebook_size - k);

			for (j=0; j < rest; j++) {

				// project remaining codebook vectors on winning codebook vector
				double p = scp(pwin, codebook_tmp[candidates[j]], cur_ssize)/nwin;

				smultiply(ctemp, pwin, p/nwin, cur_ssize);

				sub(codebook_tmp[candidates[j]], codebook_tmp[candidates[j]], ctemp, cur_ssize);

				double nn = sqrt(scp(codebook_tmp[candidates[j]], codebook_tmp[candidates[j]], cur_ssize)); 
				if(nn==0) nn=1;
				sdivide(codebook_tmp[candidates[j]], codebook_tmp[candidates[j]], nn, cur_ssize);
			}

			// every remaining codebook vector in candidates is now orthogonal with respect to cb_win

			nx = sqrt(scp(x,x, cur_ssize)); // compute norm of current residual 
		
			k++;
		}
	}
quit:
	free_array(candidates);
	free_array(temp);
	free_array(y);
	free_array(itemp);
	free_array(ranking);
	free_array(h);
	free_array(cdiff);
	free_array(ctemp);
	free_array(x);

	m_free(codebook_tmp);	

	return codebook;
}

void evaluate_current_solution(double* r_mXE, double* r_mN0, double* r_mAE,
			       int num_components, double noise_level, DataDesc* data,
          		       double** codebook, int codebook_size, double** A_orig) {

	int i=0;
	int j=0;

	int ssize = data->ssize();
	int X_columns = data->num_samples();

	double** P = m_create(X_columns, codebook_size);
	double** R = NULL;
	double* N0 = NULL;
	double* NR = NULL;
	StandardDataDesc* P_desc = new StandardDataDesc(P, codebook_size, X_columns);
	// perform orthogonal matching pursuit
	int num_codebooks = 1;
	omp (P_desc, &R, &N0, &NR, num_components, noise_level, data, &codebook, num_codebooks, codebook_size); 

	double** tP      = m_transpose(P     , X_columns, codebook_size);
	double** tA_orig = m_transpose(A_orig, X_columns, codebook_size);

	delete P_desc;

	//m_free(P);

	double mXE = 0.0;
	for (i=0;i < X_columns; i++) mXE += sqrt(scp(R[i], R[i], ssize));

	mXE /= ((double) X_columns);	
	*r_mXE = mXE;
	m_free(R);

	*r_mN0 = sum(N0, X_columns)/ ((double) X_columns);	
	double* nA_orig = double_array(codebook_size);
	double* nP      = double_array(codebook_size);

	for (i=0; i < codebook_size; i++) {
		nA_orig[i] = sqrt(scp(tA_orig[i], tA_orig[i], X_columns)); // Compute norm of each source dimension
		nP[i]      = sqrt(scp(tP[i]     , tP[i]     , X_columns)); // Compute norm of each estimated source dimension
	}
	double* temp1    = double_array(codebook_size);
	double* temp2    = double_array(X_columns);
	double** OVERLAP = m_create(codebook_size, codebook_size);

	for (i=0; i < codebook_size; i++) {
		for(j=0; j < codebook_size; j++) {
			OVERLAP[i][j] = scp(tP[i], tA_orig[j], X_columns)/(nP[i]*nA_orig[j]);
		}
	}

	double** SQR_OVERLAP = m_create(codebook_size, codebook_size);
	for (i=0; i < codebook_size; i++) multiply(SQR_OVERLAP[i], OVERLAP[i], OVERLAP[i], codebook_size);

	int* orig  = int_array(codebook_size);
	int* learn = int_array(codebook_size);

	for (i=0; i < codebook_size; i++) {

		int m_index=0;
		int entry_index=0;

		m_max(&m_index, &entry_index, SQR_OVERLAP, codebook_size, codebook_size);	

		learn[i] = m_index;
		orig[i]  = entry_index;


		for (j=0; j < codebook_size; j++) {
			SQR_OVERLAP[m_index][j]     = -1;
			SQR_OVERLAP[j][entry_index] = -1;
		}
	}

	double mAE = 0.0;
	for (i=0; i < codebook_size; i++) {

		if(OVERLAP[learn[i]][orig[i]] < 0) smultiply(tP[learn[i]], tP[learn[i]], -1, X_columns);

		sub(temp2, tP[learn[i]], tA_orig[orig[i]], X_columns);
		multiply(temp2, temp2, temp2, X_columns);
		
		mAE += sum(temp2, X_columns)/ ((double) X_columns);
	}

	mAE /= ((double) codebook_size);
	*r_mAE = mAE;

	m_free(OVERLAP);
	m_free(SQR_OVERLAP);
	free_array(orig);
	free_array(learn);

	free_array(nA_orig);
	free_array(nP);

	free_array(temp1);
	free_array(temp2);
	m_free(tP);
	m_free(tA_orig);

	free_array(N0);
	free_array(NR);
}

void omp (DataDesc* P, double*** r_R, double** r_N0, double** r_NR, int num_components, double noise_level, DataDesc* data, 
	  double*** codebooks, int num_codebooks, int codebook_size) {

	int i=0;
	int j=0;

	int X_columns = data->num_samples();
	int ssize     = data->ssize();

	double* p_temp = double_array(codebook_size);
	int* I         = int_array(codebook_size);

	double* rtemp = NULL;
	double** R    = NULL;

	if (num_components > codebook_size) {
		PRINTF("num_components > codebook_size!!! setting num_components=%i !!!", codebook_size);
		num_components = codebook_size;
	}

	if (r_R != NULL) { // return residual for each sample in X only if r_R != NULL !!!

		if (*r_R == NULL) *r_R = m_create(X_columns, ssize);
		R = *r_R;

		m_zero(R, X_columns, ssize);
	}
	else rtemp = double_array(ssize);

	double* N0 = NULL;	
	if (*r_N0 == NULL) *r_N0 = double_array(X_columns);
	N0 = *r_N0;

	double* NR = NULL;
	if (*r_NR == NULL) *r_NR = double_array(X_columns);
	NR = *r_NR;

	int* next_win   = int_array(codebook_size);
	int* candidates = int_array(codebook_size);

	double** codebook_tmp = m_create(codebook_size, ssize);	

	double* temp   = double_array(codebook_size);
	double* temp2  = double_array(ssize);
	double* winner = double_array(ssize);
	double* y      = double_array(codebook_size);

	double* rcur = rtemp;

	int next_win_len=0;
	for (j=0; j < X_columns; j++) {
		
		if(R) rcur=R[j];

		int* active_dims = NULL;
		int cur_ssize = data->CopySample(rcur, &active_dims, j);

		for(int i=0; i < codebook_size; i++) candidates[i] = i;

		int cc = (num_codebooks > 1) ? j : 0;
		
		if (active_dims==NULL) m_copy (codebook_tmp, codebooks[cc], codebook_size, ssize);
		else {
			for(int i=0; i < codebook_size; i++) get_i(codebook_tmp[i], codebooks[cc][i], active_dims, cur_ssize);
		}

		next_win_len=0;

		memset(next_win, 0, codebook_size*sizeof(int));
		memset(p_temp  , 0, codebook_size*sizeof(double));

		omp_rec (num_components, p_temp, rcur, next_win, noise_level, codebook_tmp, cur_ssize, codebook_size, 
			 candidates, y, temp, temp2, winner, &next_win_len, codebook_size);

		int I_len = find_neq(I, p_temp, 0, codebook_size);
		//PRINTF("I_len: %i\n", I_len);

		//print_array("p_temp: ", p_temp, codebook_size);

//		P->PutSample(p_temp, I, I_len, j);

		P->PutSample(p_temp, NULL, I_len, j);

		N0[j] = next_win_len;			// return zero norm of coefficients
		NR[j] = sqrt(scp(rcur, rcur, cur_ssize));   // return norm of residual
	}

	if(rtemp) free_array(rtemp);
	
	free_array(p_temp);
	free_array(I);

	free_array(next_win);
	free_array(candidates);
	free_array(temp);
	free_array(temp2);
	free_array(winner);
	free_array(y);
	m_free(codebook_tmp);
}

void omp_rec (int k, double* p, double* r, int* next_win, double noise_level, 
              double** codebook, int ssize, int codebook_size, int* candidates, 
              double* y, double* temp,double* temp2, double* winner, int* next_win_len, int cand_len) {

	int j=0;

	// calculate distance of the residual to all codebook vectors  y = w*x
	for (j=0; j < cand_len; j++) y[j] = scp(r, codebook[candidates[j]], ssize);

	multiply(temp , y   , y , cand_len); // temp = -(y^2)
	smultiply(temp, temp, -1, cand_len); 

	int i_min      = dmin(temp, cand_len);
	int win        = candidates[i_min];
	double y_win   = y[i_min];
	memcpy(winner, codebook[win], ssize*sizeof(double));

	// remove projection of residual to winner from residual
	smultiply(temp2, winner, y_win, ssize);
	sub (r,r,temp2, ssize);

	double* p_to_win = double_array(codebook_size);
	for(j=0; j < codebook_size; j++) {
		// compute projection of codebook to winning codebook_vector
		p_to_win[j] = scp(codebook[j], winner, ssize);
		smultiply(temp2, winner, p_to_win[j], ssize);
		// remove projection from codebook
		sub(codebook[j], codebook[j], temp2, ssize);
	}

	// remove winner from the list of possible candidates in subsequent iterations
	for (j=i_min; j < cand_len - 1; j++)candidates[j]=candidates[j+1];
	cand_len -=1;

	// compute norm of current residual
	double nr = sqrt(scp(r,r,ssize));

	if ((k > 1) && (nr > noise_level)) {
		double* n = double_array(codebook_size);
		memset(n, 0, codebook_size*sizeof(double));

		// set norm of remaining candidates to 1
		for (j=0; j < cand_len; j++) {
			n[candidates[j]] = sqrt(scp(codebook[candidates[j]], codebook[candidates[j]], ssize));

			if (n[candidates[j]] == 0) n[candidates[j]] = 1.0;

			sdivide(codebook[candidates[j]], codebook[candidates[j]], n[candidates[j]], ssize);
		}
		omp_rec (k-1, p, r, next_win, noise_level, codebook, ssize, codebook_size, 
			 candidates, y, temp, temp2, winner, next_win_len,cand_len);

		for (j=0; j < *next_win_len ; j++) {
			p[win] = p[win] - p[next_win[j]]*p_to_win[next_win[j]]/n[next_win[j]];
			p[next_win[j]] = p[next_win[j]]/n[next_win[j]];
		}
		p[win] = p[win] + y_win;

		next_win[*next_win_len] = win;

		(*next_win_len)++;

		free_array(n);
	}
	else {
		memset(p, 0, codebook_size*sizeof(double));
		p[win]  = y_win;
		next_win[*next_win_len] = win;
		(*next_win_len)++;
	}
	free_array(p_to_win);
}
