/*
(Soft)MaxMinOver learning algorithm written by Kai Labusch
Copyright (C) 2004 Kai Labusch

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Contact:
labusch@inb.uni-luebeck.de
*/

#include "utils.h"
#include <stdlib.h>

#include <cmath>
#include <cstring>

using namespace std;

#define CHECK_ALIGN(p,line,file) if(!is_aligned((void*)p)) PRINTF("Wrong alignment !!!! file : %s, line: %i\n", file, line);

void print_array (const char* text, double* array, int length) {

	PRINTF ("%s   [",text);
	int i=0;
	const char* sep = "";
	for (i=0; i < length; i++) {
		PRINTF ("%s%f", sep, array[i]);
		sep = ",";
	}
	PRINTF ("]\n");
}

void iprint_array (const char* text, int* array, int length) {
	PRINTF ("%s   [", text);
	int i=0;
	const char* sep= "";
	for (i=0; i < length; i++) {
		PRINTF ("%s%i", sep, array[i]);
		sep = ",";
	}
	PRINTF ("]\n");
}

/*---------*/

void assign (double* dest, double val, int length) {

	double* limit = dest + length;
	while (dest < limit) *dest++ = val;
}

void assign_i(double* dest, int* I, double val, int length) {
	int i = 0;

	for (i=0; i < length; i++) dest[I[i]] = val;
}

/*---------*/

void iassign (int* dest, int val, int length) {
	int i = 0;

	for (i=0; i < length; i++) dest[i] = val;
}

void iassign_i(int* dest, int* I, int val, int length) {
	int i = 0;

	for (i=0; i < length; i++) dest[I[i]] = val;
}

/*---------*/

double scp (double* fac1, double* fac2, int length) {

	double* limit = fac1 + length;
	double r = 0;
#ifdef __USE_SSE__
	CHECK_ALIGN(fac1, __LINE__, __FILE__);
	CHECK_ALIGN(fac2, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%esi), %%xmm1\n\t"
	"0:	\n\t"
	"dec %%eax\n\t"
	"jl 1f\n\t"
	"movapd (%%ecx),%%xmm0\n\t"
	"mulpd (%%edx), %%xmm0\n\t"
	"addpd %%xmm0,%%xmm1\n\t"
	"add $16,%%ecx\n\t"
	"add $16,%%edx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm1,%%xmm2\n\t"
	"unpckhpd %%xmm2,%%xmm2\n\t"
	"addsd %%xmm2,%%xmm1\n\t"
	"movsd %%xmm1,(%%esi)\n\t"
	: "=c" (fac1) , "=d" (fac2)  
	: "a" (cBlocks), "S" (&r) , "c" (fac1) , "d" (fac2)
	: "%xmm0" , "%xmm1" , "%xmm2" , "memory"
	);
#endif

#ifdef __USE_SSE64__
	CHECK_ALIGN(fac1, __LINE__, __FILE__);
	CHECK_ALIGN(fac2, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%rsi), %%xmm1\n\t"
	"0:	\n\t"
	"dec %%eax\n\t"
	"jl 1f\n\t"
	"movapd (%%rcx),%%xmm0\n\t"
	"mulpd (%%rdx), %%xmm0\n\t"
	"addpd %%xmm0,%%xmm1\n\t"
	"add $16,%%rcx\n\t"
	"add $16,%%rdx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm1,%%xmm2\n\t"
	"unpckhpd %%xmm2,%%xmm2\n\t"
	"addsd %%xmm2,%%xmm1\n\t"
	"movsd %%xmm1,(%%rsi)\n\t"
	: "=c" (fac1) , "=d" (fac2)  
	: "a" (cBlocks), "S" (&r) , "c" (fac1) , "d" (fac2)
	: "%xmm0" , "%xmm1" , "%xmm2" , "memory"
	);
#endif

	while(fac1 < limit) r += *fac1++ * *fac2++;
		
	return r;
}

double diff_scp(double* x1, double* x2, int length) {

	double* limit = x1 + length;
	double r = 0;
	double t = 0;
#ifdef __USE_SSE__

	CHECK_ALIGN(x1, __LINE__, __FILE__);
	CHECK_ALIGN(x2, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%esi), %%xmm1\n\t"
	"0:	\n\t"
	"dec %%eax\n\t"
	"jl 1f\n\t"
	"movapd (%%ecx),%%xmm0\n\t"
	"subpd (%%edx),%%xmm0\n\t"
	"mulpd  %%xmm0,%%xmm0\n\t"
	"addpd  %%xmm0,%%xmm1\n\t"
	"add $16,%%ecx\n\t"
	"add $16,%%edx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm1,%%xmm2\n\t"
	"unpckhpd %%xmm2,%%xmm2\n\t"
	"addsd %%xmm2,%%xmm1\n\t"
	"movsd %%xmm1,(%%esi)\n\t"
	: "=c" (x1) , "=d" (x2)  
	: "a" (cBlocks), "S" (&r) , "c" (x1) , "d" (x2)
	: "%xmm0" , "%xmm1" , "%xmm2" , "memory"
	);
#endif

#ifdef __USE_SSE64__

	CHECK_ALIGN(x1, __LINE__, __FILE__);
	CHECK_ALIGN(x2, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%rsi), %%xmm1\n\t"
	"0:	\n\t"
	"dec %%rax\n\t"
	"jl 1f\n\t"
	"movapd (%%rcx),%%xmm0\n\t"
	"subpd (%%rdx),%%xmm0\n\t"
	"mulpd  %%xmm0,%%xmm0\n\t"
	"addpd  %%xmm0,%%xmm1\n\t"
	"add $16,%%rcx\n\t"
	"add $16,%%rdx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm1,%%xmm2\n\t"
	"unpckhpd %%xmm2,%%xmm2\n\t"
	"addsd %%xmm2,%%xmm1\n\t"
	"movsd %%xmm1,(%%rsi)\n\t"
	: "=c" (x1) , "=d" (x2)  
	: "a" (cBlocks), "S" (&r) , "c" (x1) , "d" (x2)
	: "%xmm0" , "%xmm1" , "%xmm2" , "memory"
	);
#endif

	while(x1 < limit) {
		t = *x1++ - *x2++;
		r += t*t;
	}
		
	return r;
}

double l0_diff_scp(double* x1, double* x2, int length) {

	double* limit = x1 + length;
	double r = 0;

	while(x1 < limit) {
		if (*x1++ != *x2++) r++;
	}
		
	return r;
}


void multiply (double* product, double* fac1, double* fac2, int length) {

	double* limit = product + length;

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	int cBlocks = length/2;

	CHECK_ALIGN(fac1, __LINE__, __FILE__);
	CHECK_ALIGN(fac2, __LINE__, __FILE__);
	CHECK_ALIGN(product, __LINE__, __FILE__);

	while (cBlocks > 0) {
		__asm__ __volatile__ (
		"movapd (%1),%%xmm0\n\t"
		"mulpd (%2), %%xmm0\n\t"
		"movapd %%xmm0,(%0)\n\t"
		:
		: "r" (product) , "r" (fac1) , "r" (fac2)
		: "%xmm0" , "%xmm1"
		);

		cBlocks--;
		product+=2;
		fac1+=2;
		fac2+=2;
	}
#endif
	while(product< limit) *product++ = *fac1++ * *fac2++;

}

void multiply_ip (double* factor1, double* factor2, int length) {
	int i = 0;

	for (i=0; i < length; i++) factor1[i] = factor1[i] * factor2[i];
}

void multiply_i(double* product, double* factor1, double* factor2, int* I, int length) {
	int i = 0;

	for (i=0; i < length; i++) product[I[i]] = factor1[I[i]] * factor2[I[i]];
}

void multiply_ip_i(double* factor1, double* factor2, int* I, int length) {
	int i = 0;

	for (i=0; i < length; i++) factor1[I[i]] = factor1[I[i]] * factor2[I[i]];
}

void multiply_shr(double* product, double* factor1, double* factor2, int* I, int length) {
	int i = 0;

	for (i=0; i < length; i++) product[i] = factor1[I[i]] * factor2[I[i]];
}

/*----------*/

void divide (double* quotient, double* numerator, double* denominator, int length) {
	int i = 0;
	
	for (i=0; i < length; i++) quotient[i] = numerator[i] / denominator[i];
}

void divide_ip (double* numerator, double* denominator, int length) {
	int i = 0;
	
	for (i=0; i < length; i++) numerator[i] = numerator[i] / denominator[i];
}

void divide_i(double* quotient, double* numerator, double* denominator, int* I, int length) {
	int i = 0;
	
	for (i=0; i < length; i++) quotient[I[i]] = numerator[I[i]] / denominator[I[i]];
}

void divide_ip_i(double* numerator, double* denominator, int* I, int length) {
	int i = 0;
	
	for (i=0; i < length; i++) numerator[I[i]] = numerator[I[i]] / denominator[I[i]];
}

void divide_shr(double* quotient, double* numerator, double* denominator, int* I, int length) {
	int i = 0;
	
	for (i=0; i < length; i++) quotient[i] = numerator[I[i]] / denominator[I[i]];
}

/*---------*/

void add (double* sum, double* s1, double* s2, int length) {
	double* limit = sum + length;

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	int cBlocks = length/2;

	CHECK_ALIGN(s1, __LINE__, __FILE__);
	CHECK_ALIGN(s2, __LINE__, __FILE__);
	CHECK_ALIGN(sum, __LINE__, __FILE__);

	while (cBlocks > 0) {
		__asm__ __volatile__ (
		"movapd (%1),%%xmm0\n\t"
		"addpd (%2), %%xmm0\n\t"
		"movapd %%xmm0,(%0)\n\t"
		:
		: "r" (sum) , "r" (s1) , "r" (s2)
		: "%xmm0" , "%xmm1"
		);

		cBlocks--;
		sum+=2;
		s1+=2;
		s2+=2;
	}
#endif
	while(sum < limit) *sum++ = *s1++ + *s2++;
}

void add_ip (double* s1, double* s2, int length) {
	int i = 0;

	for (i=0; i < length; i++) s1[i] = s1[i] + s2[i];
}

void add_i(double* sum, double* s1, double* s2, int* I, int length) {
	int i = 0;

	for (i=0; i < length; i++) sum[I[i]] = s1[I[i]] + s2[I[i]];
}

void add_ip_i(double* s1, double* s2, int* I, int length) {
/*	int i = 0;

	for (i=0; i < length; i++) s1[I[i]] = s1[I[i]] + s2[I[i]];*/

	int* limit = I + length;

	while(I < limit - 4) {
		s1[*(I+0)] = s1[*(I+0)] + s2[*(I+0)];
		s1[*(I+1)] = s1[*(I+1)] + s2[*(I+1)];
		s1[*(I+2)] = s1[*(I+2)] + s2[*(I+2)];
		s1[*(I+3)] = s1[*(I+3)] + s2[*(I+3)];

		I+=4;
	}
	
	while(I < limit) s1[*I++] = s1[*I] + s2[*I];

}


void add_shr(double* sum, double* s1, double* s2, int* I, int length) {
	int i = 0;

	for (i=0; i < length; i++) sum[i] = s1[I[i]] + s2[I[i]];
}

/*---------*/


void sub (double* diff, double* s1, double* s2, int length) {
	double* limit = diff + length;

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	int cBlocks = length/2;

	CHECK_ALIGN(s1, __LINE__, __FILE__);
	CHECK_ALIGN(s2, __LINE__, __FILE__);
	CHECK_ALIGN(diff, __LINE__, __FILE__);
	
	while (cBlocks > 0) {
		__asm__ __volatile__ (
		"movapd (%1),%%xmm0\n\t"
		"subpd (%2), %%xmm0\n\t"
		"movapd %%xmm0,(%0)\n\t"
		:
		: "r" (diff) , "r" (s1) , "r" (s2)
		: "%xmm0" , "%xmm1"
		);

		cBlocks--;
		diff+=2;
		s1+=2;
		s2+=2;
	}
#endif
	while(diff < limit) *diff++ = *s1++ - *s2++;
}

void sub_ip (double* sum1, double* sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[i] = sum1[i] - sum2[i];
}

void sub_i(double* diff, double* sum1, double* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[I[i]] = sum1[I[i]] - sum2[I[i]];
}

void sub_ip_i(double* sum1, double* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[I[i]] = sum1[I[i]] - sum2[I[i]];
}

void sub_shr(double* diff, double* sum1, double* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[I[i]] - sum2[I[i]];
}

/*-------*/

void isub (int* diff, int* sum1, int* sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[i] - sum2[i];
}

void isub_ip (int* sum1, int* sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[i] = sum1[i] - sum2[i];
}

void isub_i(int* diff, int* sum1, int* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[I[i]] = sum1[I[i]] - sum2[I[i]];
}

void isub_ip_i(int* sum1, int* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[I[i]] = sum1[I[i]] - sum2[I[i]];
}

void isub_shr(int* diff, int* sum1, int* sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[I[i]] - sum2[I[i]];
}

/*--------*/

void absol (double* xabs, double* x, int length) {
	int i=0;

	for (i=0; i < length; i++) xabs[i] = fabs(x[i]);
}


void absol_i(double* xabs, double* x, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) xabs[I[i]] = fabs(x[I[i]]);
}

/*-----------*/

void smultiply(double* product, double* fac1, double fac2, int length) {

	double* limit = product + length;

#ifdef __USE_SSE__

	CHECK_ALIGN(fac1, __LINE__, __FILE__);
	CHECK_ALIGN(product, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%edx), %%xmm0\n\t"
	"unpcklpd %%xmm0, %%xmm0\n\t"	
	"0:	\n\t"
	"dec %%eax\n\t"
	"jl 1f\n\t"
	"movapd (%%ecx),%%xmm1\n\t"
	"mulpd %%xmm0, %%xmm1\n\t"
	"movapd %%xmm1,(%%esi)\n\t"
	"add $16,%%esi\n\t"
	"add $16,%%ecx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	: "=S" (product) , "=c" (fac1)  
	: "a" (cBlocks), "S" (product) , "c" (fac1) , "d" (&fac2)
	: "%xmm0" , "%xmm1"
	);
#endif

#ifdef __USE_SSE64__

	CHECK_ALIGN(fac1, __LINE__, __FILE__);
	CHECK_ALIGN(product, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%rdx), %%xmm0\n\t"
	"unpcklpd %%xmm0, %%xmm0\n\t"	
	"0:	\n\t"
	"dec %%rax\n\t"
	"jl 1f\n\t"
	"movapd (%%rcx),%%xmm1\n\t"
	"mulpd %%xmm0, %%xmm1\n\t"
	"movapd %%xmm1,(%%rsi)\n\t"
	"add $16,%%rsi\n\t"
	"add $16,%%rcx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	: "=S" (product) , "=c" (fac1)  
	: "a" (cBlocks), "S" (product) , "c" (fac1) , "d" (&fac2)
	: "%xmm0" , "%xmm1"
	);
#endif


	while(product < limit) *product++ = *fac1++ * fac2;

}

void smultiply_ip(double* fac1, double fac2, int length) {
	//int i=0;
	//for (i=0; i < length; i++) fac1[i] = fac1[i]*fac2;

	smultiply(fac1, fac1, fac2, length);
}

void smultiply_i(double* product, double* fact1, double fac2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) product[I[i]] = fact1[I[i]]*fac2;
}

void smultiply_ip_i(double* fact1, double fac2, int* I, int length) {

	int* limit = I + length;

	while(I < limit) fact1[*I++] *= fac2;
}

void smultiply_shr(double* product, double* fact1, double fac2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) product[i] = fact1[I[i]]*fac2;
}

/*------------*/

void sdivide(double* quotient, double* numerator, double denominator, int length) {
/*	int i=0;

	for (i=0; i < length; i++) quotient[i] = numerator[i] / denominator;*/

	smultiply(quotient, numerator, (1.0/denominator), length);
}

void sdivide_ip(double* numerator, double denominator, int length) {
	int i=0;

	for (i=0; i < length; i++) numerator[i] = numerator[i] / denominator;
}

void sdivide_i(double* quotient, double* numerator, double denominator, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) quotient[I[i]] = numerator[I[i]] / denominator;
}

void sdivide_ip_i(double* numerator, double denominator, int* I, int length) {

	int i=0;

	for (i=0; i < length; i++) numerator[I[i]] = numerator[I[i]] / denominator;
}

void sdivide_shr(double* quotient, double* numerator, double denominator, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) quotient[i] = numerator[I[i]] / denominator;
}

/*------------*/

void sadd(double* sum, double* s1, double s2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum[i] = s1[i] + s2;
}

void psadd(double** sum, double** s1, int offset, int length) {
	double** limit = sum + length;
	
	while (sum < limit) *sum = *sum++ + offset;
}

void isadd(int* sum, int* s1, int s2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum[i] = s1[i] + s2;
}


void sadd_ip(double* s1, double s2, int length) {
	int i=0;

	for (i=0; i < length; i++) s1[i] = s1[i] + s2;
}

void sadd_i(double* sum, double* s1, double s2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) sum[I[i]] = s1[I[i]] + s2;
}

void isadd_i(int* sum, int* s1, int s2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) sum[I[i]] = s1[I[i]] + s2;
}


void sadd_ip_i(double* s1, double s2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) s1[I[i]] = s1[I[i]] + s2;
}

void isadd_ip_i(int* s1, int s2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) s1[I[i]] = s1[I[i]] + s2;
}


void sadd_shr(double* sum, double* s1, double s2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) sum[i] = s1[I[i]] + s2;
}

/*-------------*/

void ssub(double* diff, double* sum1, double sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[i] - sum2;
}

void issub(int* diff, int* sum1, int sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[i] - sum2;
}

void ssub_ip(double* sum1, double sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[i] = sum1[i] - sum2;
}

void ssub_i(double* diff, double* sum1, double sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[I[i]] = sum1[I[i]] - sum2;
}

void ssub_ip_i(double* sum1, double sum2, int* I, int length) {

	int* limit = I + length;

	while(I < limit) sum1[*I++] -= sum2;
}


void ssub_shr(double* diff, double* sum1, double sum2, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) diff[i] = sum1[I[i]] - sum2;
}

/*--------------*/

void issub_ip(int* sum1, int sum2, int length) {
	int i=0;

	for (i=0; i < length; i++) sum1[i] = sum1[i] - sum2;
}

/*--------------*/


void exponential10 (double* dest, double* x, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[i] = exp(log(10.0)*x[i]);
}

void exponential (double* dest, double* x, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[i] = exp(x[i]);
}

void exponential_i(double* dest, double* x, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[I[i]] = exp(x[I[i]]);	
}

/*--------------*/

void power (double* dest, double* x, double y, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[i] = pow(x[i], y);
}

void power2 (double* dest, double x, double* y, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[i] = pow(x, y[i]);
}


void power_i(double* dest, double* x, double y, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[I[i]] = pow(x[I[i]], y);
}

/*---------------*/

void squarert (double* dest, double* x, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[i] = sqrt(x[i]);
}

void squarert_i (double* dest, double* x, int* I, int length) {
	int i=0;

	for (i=0; i < length; i++) dest[I[i]] = sqrt(x[I[i]]);
}

/*---------------*/

double mean (double* x, int length) {
	int i=0;
	double m = 0.0;

	for (i=0; i < length; i++) m = m + x[i];

	return m / (double) length;
}


double mean_i (double* x, int* I, int length) {
	int i=0;
	double m = 0.0;

	for (i=0; i < length; i++) m = m + x[I[i]];

	return m / (double) length;
}

double var (double* x, int length) {
	double* temp = double_array(length);

	double me = mean(x, length);
	ssub (temp, x, me, length);
	multiply(temp, temp, temp, length);

	double v = sum(temp, length);

	free_array(temp);

	return (1/(((double)length) - 1))*v;	
}

/*---------------*/

double prod (double* x, int length) {
	double r=1;

	int i=0;
	for (i=0; i < length; i++) r=r*x[i];

	return r;
}

int prod_i (int* x, int length) {
	int r=1;

	int i=0;
	for (i=0; i < length; i++) r=r*x[i];

	return r;
}

double sum (double* x, int length) {

	double* limit = x + length;
	double s = 0;
#ifdef __USE_SSE__

	CHECK_ALIGN(x, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%esi), %%xmm0\n\t"
	"0:	\n\t"
	"dec %%eax\n\t"
	"jl 1f\n\t"
	"addpd (%%ecx),%%xmm0\n\t"
	"add $16,%%ecx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm0,%%xmm1\n\t"
	"unpckhpd %%xmm1,%%xmm1\n\t"
	"addsd %%xmm1,%%xmm0\n\t"
	"movsd %%xmm0,(%%esi)\n\t"
	: "=c" (x)
	: "a" (cBlocks), "S" (&s) , "c" (x)
	: "%xmm0" , "%xmm1" , "memory"
	);
#endif
#ifdef __USE_SSE64__

	CHECK_ALIGN(x, __LINE__, __FILE__);

	int cBlocks = length/2;

	__asm__ __volatile__ (
	"movsd (%%rsi), %%xmm0\n\t"
	"0:	\n\t"
	"dec %%rax\n\t"
	"jl 1f\n\t"
	"addpd (%%rcx),%%xmm0\n\t"
	"add $16,%%rcx\n\t"
	"jmp 0b\n\t"
	"1:	\n\t"
	"movapd %%xmm0,%%xmm1\n\t"
	"unpckhpd %%xmm1,%%xmm1\n\t"
	"addsd %%xmm1,%%xmm0\n\t"
	"movsd %%xmm0,(%%rsi)\n\t"
	: "=c" (x)
	: "a" (cBlocks), "S" (&s) , "c" (x)
	: "%xmm0" , "%xmm1" , "memory"
	);
#endif
	while(x < limit) s += *x++;
		
	return s;
}


double sum_i (double* x, int* I, int length) {
	int i=0;
	double s = 0;

	for (i=0; i < length; i++) s = s + x[I[i]];

	return s;
}

/*---------------*/

int isum (int* x, int length) {
	int i=0;
	int s = 0;

	for (i=0; i < length; i++) s = s + x[i];

	return s;
}


int isum_i (int* x, int* I, int length) {
	int i=0;
	int s = 0;

	for (i=0; i < length; i++) s = s + x[I[i]];

	return s;
}

/*---------------*/


void rand_array (int* dest, int min, int max, int dest_length) {
	int i=0;
	double r = max - min;

	for (i=0; i < dest_length; i++) {
		dest[i] = min + (int) ((r*(double)rand())/(RAND_MAX + 1.0));
	}
}

void drand_array (double* dest, double min, double max, int dest_length) {
	int i=0;
	double r = max - min;

	for (i=0; i < dest_length; i++) {
		dest[i] = min + (r*(double)rand())/(double)(RAND_MAX + 1.0);
	}
}


int uv_hash_func (int* hash_func_params, int* I, int I_length, int max) {
	int i=0;

	double h = 0;

	for(i=0; i < I_length; i++) {
		h += hash_func_params[i] * I[i];
	}

	h = h - floor(h / max)*max;

	return (int) h;

}

/*----------*/

int dmax (double* x, int length) {
	if (length > 0) {
		int mi = 0;
		int i=0;

		for (i=0; i < length; i++) mi = (x[mi] < x[i]) ? i : mi;

		return mi;
	}
	else return -1;
}


int dmin (double* x, int length) {
	if (length > 0) {
		int mi = 0;
		int i=0;
		for (i=0; i < length; i++) mi = (x[mi] > x[i]) ? i : mi;

		return mi;
	}
	else return -1;
}

/*----------*/

int imax (int* x, int length) {
	if (length > 0) {
		int mi = 0;
		int i=0;

		for (i=0; i < length; i++) mi = (x[mi] < x[i]) ? i : mi;

		return mi;
	}
	else return -1;
}

/*-----------*/

int imin (int* x, int length) {
	if (length > 0) {
		int mi = 0;
		int i=0;

		for (i=0; i < length; i++) mi = (x[mi] > x[i]) ? i : mi;

		return mi;
	}
	else return -1;
}


int imin_i (int* x, int* I, int length) {
	if (length > 0) {
		int mi = 0;
		int i=0;

		mi = I[0];
		for (i=0; i < length; i++) mi = (x[mi] > x[I[i]]) ? i : mi;

		return mi;
	}
	else return -1;
}

void iceil(int* I, double* x, int x_length ) {
	int i=0;

	for (i=0; i < x_length; i++) {
		I[i] = (int) ceil(x[i]);
	}
}

/*----------------------------------------------------------------------------*/

int dcompare (const void* pa, const void* pb) {
	double a = *((double*) pa);	
	double b = *((double*) pb);

	if (a < b) return -1;
	if (a > b) return 1;

	return 0;
}

int icompare (const void* pa, const void* pb) {
	int a = *((int*) pa);	
	int b = *((int*) pb);

	if (a < b) return -1;
	if (a > b) return 1;

	return 0;
}

int unique(double* un, double* x, int length, int* I) {

	double* temp = double_array(length);
	memcpy(temp, x, sizeof(double)*length);

	int* itemp = int_array(length);
	for (int i=0; i < length; i++) itemp[i] = i;

	isort(temp, itemp, length);

	int pos = 0;
	for (int i=0; i < length; i++) {
		if (i==0) { 
			*un = *temp;
			if (I) I[0]= itemp[0];
		}
		else if (temp[i] != un[pos]){
			pos++;
			un[pos]      = temp[i];
	
			if(I) I[pos] = itemp[i];
		}
	}

	free_array (temp);
	free_array (itemp);

	return (length > 0) ? pos + 1 : 0;
}

int iunique(int* un, int* x, int length) {

	int * temp = int_array(length);
	memcpy(temp, x, sizeof(int)*length);

	qsort((void*) temp, length, sizeof(int), &icompare);

	int i=0;
	int pos = 0;
	for (i=0; i < length; i++) {
		if (i==0) *un = *temp;
		else if (temp[i] != un[pos]){
			pos++;
			un[pos] = temp[i];
		}
	}

	free_array (temp);

	return (length > 0) ? pos + 1 : 0;
}

void isort(double* x, int* index, int length) {
	int i=0;
	double* sort_array = double_array(length*2);
	memset(sort_array, 0, length*2*sizeof(double));

	for (i=0;i < length*2;i+=2) {
		sort_array[i]  = x[i/2];
		if (index) sort_array[i+1]= (double) index[i/2];
	}

	qsort((void*) sort_array, length, sizeof(double)*2, &dcompare);

	for (i=0;i < length*2;i+=2) { // extract indices
		x[i/2]  = sort_array[i];	
		if(index) index[i/2] = (int) sort_array[i+1];
	}

	free_array(sort_array);
}

void i_isort(int* x, int* index, int length) {
	int i=0;
	int* sort_array = int_array(length*2);
	memset(sort_array, 0, length*2*sizeof(int));

	for (i=0;i < length*2;i+=2) {
		sort_array[i]             = x[i/2];
		if (index) sort_array[i+1]= index[i/2];
	}

	qsort((void*) sort_array, length, sizeof(int)*2, &icompare);

	for (i=0;i < length*2;i+=2) { // extract indices
		x[i/2]  = sort_array[i];	
		if(index) index[i/2] = sort_array[i+1];
	}

	free_array(sort_array);
}


int get_urandom_entries (int* dest, int* rest, int* I, int I_length, int count) {
	if (count > I_length) {
		PRINTF("I_length < count !!! , I_length: %i, count: %i\n", I_length, count);
	
		return -1;
	}	
	
	int rest_length = I_length;
	int dest_length = 0;	

	memcpy(rest, I, I_length*sizeof(int));

	int* itemp = int_array(I_length);
	int  itemp_length = 0;

	while (dest_length < count) {

		if (rest_length > 1) {
			rand_array (itemp, 0, rest_length - 1, count - dest_length);
			iget_i(dest + dest_length, rest, itemp, count - dest_length);
		
			dest_length = iunique(itemp, dest, dest_length + count - dest_length); 
			memcpy(dest, itemp, dest_length*sizeof(int));

			set_diff (dest, I, dest_length, I_length, itemp, rest, &itemp_length, &rest_length);
		}
		else {
			dest[dest_length++] = rest[0];
			rest_length = 0;
		}
	}
	free_array(itemp);

	return rest_length;
}

void randperm(int* p, int len) {
	for (int i=0; i < len; i++) p[i] = i;

	double* temp = double_array(len);

	drand_array (temp, 0, RAND_MAX, len);

	isort(temp, p, len);

	free_array(temp);
}

/*----------------------------------------------------------------------------*/

void chartoint3(char* dest, char* src, int length) {

	unsigned char* limit = (unsigned char*) (dest + (length*3 + 7)/8 ); /*ceil*/
	unsigned char* pos   = (unsigned char*) dest;
	int src_bits  = 3;
	int dest_bits = 8; 
	char* src_limit = src + length;
	unsigned char cur_src  = ((unsigned char) *src++) << 5;
	unsigned char cur_dest = 0;

	int counter =0;

	while (pos < limit) {

		cur_dest |= cur_src >> (8 - dest_bits);	

		if (dest_bits >= src_bits) {
			dest_bits -= src_bits;
			if (src < src_limit) cur_src  = ((unsigned char) *src++) << 5;
			else cur_src = 0;
			src_bits = 3;
		}
		else {
			cur_src = cur_src << dest_bits;
			src_bits -= dest_bits;

			*pos++    = cur_dest;
			dest_bits = 8;
			cur_dest  = 0;
		}
	}
}

void int3tochar(char* dest, char* src, int length) {
	
	unsigned char* limit = (unsigned char*) (src + (length*3+7)/8); /* ceil */
	unsigned char* pos = (unsigned char*) src;
	char* dest_limit = dest + length;
	int src_bits  = 8;
	int dest_bits = 0; 
	unsigned char cur_src  = *pos++;
	unsigned char cur_dest = 0;
	int use_bits = 0;

	int counter = 0;

	while ((pos < limit) || (src_bits >= 3)) {
		if (src_bits >= 3) {
			cur_dest |= ((cur_src & (224 << dest_bits)) >> dest_bits);	
			cur_src  = cur_src << (3 - dest_bits);

			src_bits  -= 3-dest_bits;
			dest_bits += 3-dest_bits;
		}
		else {
			use_bits = (src_bits > 3 - dest_bits) ? 3 - dest_bits : src_bits;

			cur_dest  |= ((cur_src & (224 << (3 - use_bits))) >> dest_bits);
			dest_bits += use_bits;
			src_bits -= use_bits;

			if (!src_bits) {
				cur_src  = *pos++;
				src_bits = 8;
			}
		}	
	
		if (dest_bits == 3) {
			if (dest < dest_limit) *dest++ = (char) (cur_dest >> 5);
			
			cur_dest  = 0;	
			dest_bits = 0;
		}
	}
}

/*----------------------------------------------------------------------------*/

void bin2char(char* dest, char* src, int length) {
	unsigned char* limit = (unsigned char*) (src + length/8); /* floor */
	unsigned char* cur   = (unsigned char*) src;
	unsigned char* tdest = (unsigned char*) dest;

	while (cur < limit) {
		*tdest++ = (*cur & (128 >> 0)) >> 7;
		*tdest++ = (*cur & (128 >> 1)) >> 6;
		*tdest++ = (*cur & (128 >> 2)) >> 5;
		*tdest++ = (*cur & (128 >> 3)) >> 4;
		*tdest++ = (*cur & (128 >> 4)) >> 3;
		*tdest++ = (*cur & (128 >> 5)) >> 2;
		*tdest++ = (*cur & (128 >> 6)) >> 1;
		*tdest++ = (*cur & (128 >> 7)) >> 0;
		cur++;
	}
	int residual = length - (length/8)*8;
	limit = tdest;
	
	while (tdest < limit + residual) {
		*tdest = (*cur & (128 >> (tdest-limit))) ? 1 : 0;
		tdest++;
	}
}

void char2bin(char* dest, char* src, int length) {
	unsigned char* limit = (unsigned char*) (dest + length/8); /* floor */
	unsigned char* cur   = (unsigned char*) dest;
	unsigned char* tsrc  = (unsigned char*) src;

	while (cur < limit) {
		*cur = 0;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 0;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 1;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 2;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 3;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 4;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 5;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 6;
		*cur |= ((*tsrc++ != 0) ? 128 : 0) >> 7;
		cur++;
	}

	int residual = length - (length/8)*8;
	
	if (residual > 0) {
		limit = tsrc; 
		*cur = 0;
		while (tsrc < limit + residual) {
			*cur |= ((*tsrc != 0) ? 128:0) >> (tsrc - limit);
			tsrc++;
		}
	}
}

int isNAN(int* dest, double* x, int length) {

	int count = 0;
	double* limit = x + length;
	int pos = 0;

	while(x < limit) {
		if (isnan(*x)) {
			count++;
			if (dest != NULL) *dest++ = pos;

		}
		x++;
		pos++;
	}

	return count;
}

int isAN(int* dest, double* x, int length) {

	int count = 0;
	double* limit = x + length;
	int pos = 0;

	while(x < limit) {
		if (!isnan(*x)) {
			count++;
			if (dest != NULL) *dest++ = pos;

		}
		x++;
		pos++;
	}

	return count;
}


/*----------------------------------------------------------------------------*/

int find_neq (int* dest_indizes, double* x, double val, int x_length) {
	double* start = x;
	int* cur_ind = dest_indizes;

	while(x < start + x_length) {
		if (*x != val) *cur_ind++ = (x - start);	
		x++;
	}

	return (cur_ind - dest_indizes);
}

int cfind_neq (int* dest_indizes, char* x, char val, int x_length) {
	char* start = x;
	int* cur_ind = dest_indizes;

	while(x < start + x_length) {
		if (*x != val) *cur_ind++ = (x - start);	
		x++;
	}

	return (cur_ind - dest_indizes);
}

int ifind_neq (int* dest_indizes, int* x, int val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] != val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}

int find_eq  (int* dest_indizes, double* x, double val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] == val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}

int ifind_eq  (int* dest_indizes, int* x, int val, int x_length) {
	int count = 0;
	int i;

	if (dest_indizes != NULL) {
		for (i = 0; i < x_length; i++) {
			if (x[i] == val) {
				dest_indizes[count] = i;	
				count++;
			}
		}
	}
	else {
		for (i = 0; i < x_length; i++) {
			if (x[i] == val) count ++;
		}
	}

	return count;
}

int sfind_eq  (int* dest_indizes, char** strs, const char* str, int strs_length) {
	return sfind_eq  (dest_indizes, strs, (char*) str, strs_length);
}


int sfind_eq  (int* dest_indizes, char** strs, char* str, int strs_length) {
	int count = 0;
	int i;

	if (dest_indizes != NULL) {
		for (i = 0; i < strs_length; i++) {
			if (strcmp(strs[i],str) == 0) {
				dest_indizes[count] = i;	
				count++;
			}
		}
	}
	else {
		for (i = 0; i < strs_length; i++) {
			if (strcmp(strs[i],str) == 0) count++;
		}
	}

	return count;
}


int find_l (int* dest_indizes, double* x, double val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] < val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}

int find_leq (int* dest_indizes, double* x, double val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] <= val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}


int find_g (int* dest_indizes, double* x, double val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] > val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}

int find_geq (int* dest_indizes, double* x, double val, int x_length) {
	int count = 0;
	int i;

	for (i = 0; i < x_length; i++) {
		if (x[i] >= val) {
			dest_indizes[count] = i;	
			count++;
		}
	}

	return count;
}

int get_i(double* dest, double* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;

	while (indizes < limit - 4) {
		*(dest + 0) = x[*(indizes + 0)]; 
		*(dest + 1) = x[*(indizes + 1)]; 
		*(dest + 2) = x[*(indizes + 2)]; 
		*(dest + 3) = x[*(indizes + 3)]; 
		indizes+=4;
		dest+=4;
	}

	while (indizes < limit) *dest++ = x[*indizes++]; 
	
	return ind_length;
}

void set_i(double* dest, double* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;

	while (indizes < limit - 4) {
		dest[*(indizes+0)] = *(x+0);
		dest[*(indizes+1)] = *(x+1);
		dest[*(indizes+2)] = *(x+2);
		dest[*(indizes+3)] = *(x+3);

		indizes+=4;
		x+=4;
	}

	while (indizes < limit) dest[*indizes++] = *x++;
}

void copy_i(double* dest, double* x, int* dest_indizes, int* x_indizes, int ind_length) {

	int* limit = dest_indizes + ind_length;

	while (dest_indizes < limit - 4) {
		dest[*(dest_indizes+0)] = x[*(x_indizes + 0)]; 
		dest[*(dest_indizes+1)] = x[*(x_indizes + 1)]; 
		dest[*(dest_indizes+2)] = x[*(x_indizes + 2)]; 
		dest[*(dest_indizes+3)] = x[*(x_indizes + 3)]; 

		dest_indizes+=4;
		x_indizes+=4;
	}

	while (dest_indizes < limit) { 
		dest[*dest_indizes] = x[*x_indizes]; 
		dest_indizes++;
		x_indizes++;
	}
}


int iget_i(int* dest, int* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit) *dest++ = x[*indizes++];
	
	return ind_length;
}

void iset_i(int* dest, int* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit) dest[*indizes++] = *x++;
}

int cget_i(char* dest, char* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit)	*dest++ = x[*indizes++];
	
	return ind_length;
}

void cset_i(char* dest, char* x, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit)	dest[*indizes++] = *x++;
}

void pset_i(double** dest_P, double** P, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while(indizes < limit)dest_P[*indizes++] = *P++;
}

int pget_i(double** dest_P, double** P, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit)	*dest_P++ = P[*indizes++];
	
	return ind_length;
}

int cpget_i(char** dest_P, char** P, int* indizes, int ind_length) {
	int* limit = indizes + ind_length;
	while (indizes < limit)	*dest_P++ = P[*indizes++];
	
	return ind_length;
}

/*------------------------------------------------------------------------------------------------*/

int iremove_entry(int* dest, int* src, int r, int length) {

	if (length > 0) {
		int* limit = src + length;
		int pos = 0;

		while (src < limit) {
			if (pos == r) src++;
			else *dest++ = *src++;
			pos++;
		}
		return length - 1;
	}
	else return length;
}

int iremove_value(int* dest, int* src, int v, int length) {

	if (length > 0) {
		int* limit = src + length;

		while (src < limit) {
			if (*src == v) src++;
			else *dest++ = *src++;
		}
		return length - 1;
	}
	else return length;
}


/*------------------------------------------------------------------------------------------------*/

double** m_transpose(double** m, long m_size, long m_entry_size) {
	long i=0;
	long j=0;

	double** ret = m_create(m_entry_size, m_size);
	
	for (i=0; i < m_entry_size; i++)
		for (j=0; j < m_size; j++)
			(ret[i])[j] = (m[j])[i];
	return ret;
}

double** m_create (long m_size, long entry_size) {
	double** ret = (double**) pointer_array(m_size);

#if defined(__USE_SSE__) | defined(__USE_SSE64__) 
	long m_entry_size = (entry_size % 2 != 0) ? entry_size + 1 : entry_size;
	ret[0] = double_array(m_size*m_entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*m_entry_size;
#else
	ret[0] = double_array(m_size*entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*entry_size;
#endif
	return ret;
}

float** fm_create (long m_size, long entry_size) {
	float** ret = (float**) pointer_array(m_size);

#if defined(__USE_SSE__) | defined(__USE_SSE64__) 
	long m_entry_size = (entry_size % 2 != 0) ? entry_size + 1 : entry_size;
	ret[0] = float_array(m_size*m_entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*m_entry_size;
#else
	ret[0] = float_array(m_size*entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*entry_size;
#endif
	return ret;
}

int** im_create (long m_size, long entry_size) {
	int** ret = (int**) pointer_array(m_size);

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	long m_entry_size = entry_size;
	while (m_entry_size % 4 != 0)  m_entry_size += 1;

	ret[0] = int_array(m_size*m_entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*m_entry_size;
#else
	ret[0] = int_array(m_size*entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*entry_size;
#endif
	return ret;
}

char** cm_create (long m_size, long entry_size) {
	char** ret = (char**) pointer_array(m_size);

	ret[0] = char_array(m_size*entry_size);	

	long i = 0;
	for (i=0; i < m_size; i++) ret[i] = ret[0] + i*entry_size;
	
	return ret;
}

void im_copy (int** dest, int** src, long m_size, long entry_size) {

	long i=0;
	for(i=0; i < m_size; i++) memcpy(dest[i], src[i], entry_size * sizeof(int));
}

void cm_copy (char** dest, char** src, long m_size, long entry_size) {

	long i=0;
	for(i=0; i < m_size; i++) memcpy(dest[i], src[i], entry_size * sizeof(char));
}

char** strs_copy(char** strs, int n) {
	int i=0;
	char** ret = (char**) pointer_array(n);

	for (i=0; i < n; i++) {
		ret[i] = char_array(strlen(strs[i]) + 1);
		strcpy(ret[i], strs[i]);
	}

	return ret;
}

void m_copy (double** dest, double** src, long m_size, long entry_size) {

	long i=0;
	for(i=0; i < m_size; i++) memcpy(dest[i], src[i], entry_size * sizeof(double));
}

void m_free (double** m) {
	free_array (m[0]);
	free_array (m);
}

void fm_free (float** m) {
	free_array (m[0]);
	free_array (m);
}

void im_free (int** m) {
	free_array (m[0]);
	free_array (m);
}

void cm_free (char** m) {
	free_array (m[0]);
	free_array (m);
}

void m_print (const char* text, double** m, int m_size, int m_entry_size) {
	int i=0;
	int j=0;

	PRINTF("%s\n", text);

	for (i=0; i < m_size; i++) {
		for (j=0; j < m_entry_size; j++) PRINTF("%g ", m[i][j]);

		PRINTF("\n");
	}
}

double** m_concatenate (double** m1, int m1_size, double** m2, int m2_size, int entry_size) {
	double** ret = m_create (m1_size + m2_size, entry_size);
	
	int r1_size = (m1[m1_size - 1] + entry_size) - m1[0];
	int r2_size = (m2[m2_size - 1] + entry_size) - m2[0];

	int rd_size = (ret[m1_size + m2_size -1] + entry_size) - ret[0];

	if (r1_size + r2_size == rd_size) {
		memcpy(ret[0], m1[0], r1_size*sizeof(double));
		memcpy(ret[m1_size], m2[0], r2_size*sizeof(double));
	}
	else {
		int i = 0;
		for(i=0; i < m1_size; i++) memcpy(ret[i], m1[i], entry_size*sizeof(double));
		for(i=0; i < m2_size; i++) memcpy(ret[i+m1_size], m2[i], entry_size*sizeof(double));
	}

	return ret;
}

void cm_to_vec (char* v, char** m, int m_size, int entry_size) {

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	for (int i=0;i < m_size; i++) memcpy (v + i*entry_size, m[i], entry_size*sizeof(char));
#else
	memcpy(v, m[0], entry_size * m_size * sizeof(char));
#endif
}


void im_to_vec (int* v, int** m, int m_size, int entry_size) {

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	for (int i=0;i < m_size; i++) memcpy (v + i*entry_size, m[i], entry_size*sizeof(int));
#else
	memcpy(v, m[0], entry_size * m_size * sizeof(int));
#endif
}


void m_to_vec (double* v, double** m, long m_size, long entry_size) {

#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	for (long i=0;i < m_size; i++) memcpy (v+ i*entry_size, m[i], entry_size*sizeof(double));
#else
	memcpy(v, m[0], entry_size * m_size * sizeof(double));
#endif
}

void vec_to_m(double** m, double* v, long m_size, long entry_size) {
	
	for (long i=0; i < m_size; i++) memcpy(m[i], v + i*entry_size, entry_size*sizeof(double));
}

void m_zero (double** m, long m_size, long entry_size) {
	long r_size = (m[m_size - 1] + entry_size) - m[0];

	memset(m[0], 0, r_size * sizeof(double));
}

void im_zero (int** m, long m_size, long entry_size) {
	long r_size = (m[m_size - 1] + entry_size) - m[0];

	memset(m[0], 0, r_size * sizeof(int));
}

void m_assign(double** m, long m_size, long entry_size, double val) {

	long r_size = (m[m_size - 1] + entry_size) - m[0];
	assign(m[0], val, r_size);	
}

void im_assign(int** m, long m_size, long entry_size, int val) {

	long r_size = (m[m_size - 1] + entry_size) - m[0];
	iassign(m[0], val, r_size);	
}


void m_max(int* m_index, int* entry_index, double** m, int m_size, int entry_size) {

	int k=0;

	double* entry_max   = double_array(m_size);
	int* entry_indices  = int_array(m_size);

	for (k=0; k < m_size; k++) {
		entry_indices[k] = dmax(m[k], entry_size);
		entry_max[k] = m[k][entry_indices[k]];
	}

	*m_index     = dmax(entry_max, m_size);
	*entry_index = entry_indices[*m_index]; 

	free_array(entry_max);
	free_array(entry_indices);
}

void m_min(int* m_index, int* entry_index, double** m, int m_size, int entry_size) {

	int k=0;

	double* entry_min = double_array(m_size);
	int* entry_indices  = int_array(m_size);

	for (k=0; k < m_size; k++) {
		entry_indices[k] = dmin(m[k], entry_size);
		entry_min[k]     = m[k][entry_indices[k]];
	}

	*m_index     = dmin(entry_min, m_size);
	*entry_index = entry_indices[*m_index]; 

	free_array(entry_min);
	free_array(entry_indices);
}


/*---------------------------------------------------------------------------------------------*/

int bucket_sort (int* dest, int* x, int x_length, int* bucket, int bucket_size) {
	if (x_length == 1) {
		*dest = *x;
		return 1;
	}
	else {
		memset (bucket, 0, bucket_size*sizeof(int));
		int i=0;
		int j=0;
		for (i=0; i < x_length; i++) {
			bucket[x[i]]++;
		}

		for (i=0; i < bucket_size; i++) {
			if (bucket[i]) {
				dest[j] = i;
				j++;

				if (j >= x_length) break;
			}
		}
		return j;
	}
}

/*------------------------------------------------------------------------------------------------*/

void* get_aligned_mem(long size, long alignment){
#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	void* pv;
	void** ppv;
	char* pb;

	// Add extra bytes for alignment and for storing original addr
	if(alignment<sizeof(void*)) alignment=sizeof(void*);
	size+=2*alignment;

	// Allocate memory and find first byte in block that is aligned
	pv=MALLOC(size);
	if (pv == NULL) PRINTF("MALLOC failed! size: %li.\n", size);
	

	pb=(char*)pv;
	pb+=sizeof(void*);
	while( ((long)pb)%alignment!=0) pb++;

	// Remember original non-aligned address
	ppv=(void**)pb;
	ppv[-1]=pv;

	return pb;
#else
	return MALLOC(size);
#endif
}

void free_aligned_mem(void* p) {
#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	void **ppv;

	if(p !=NULL) { // Free original non-aligned address
		ppv=(void**) p;
		FREE(ppv[-1]);
	}
#else
	FREE(p);
#endif
}

int is_aligned (void* p) {
	return ( ((long)p % 16) == 0 );
}

int align_data (void** pp, long size) {
#if defined(__USE_SSE__) | defined(__USE_SSE64__)
	if (is_aligned(*pp)) return 1;
	else {
		PRINTF ("align data\n");
		void* p = char_array(size);
		memcpy(p, *pp, size);
		*pp = p;
		return 0;
	}
#else
	return 1;
#endif
} 

int* int_array (long length) {
	return (int*) get_aligned_mem(length*sizeof(int), 16);
}

long* long_array (long length) {
	return (long*) get_aligned_mem(length*sizeof(long), 16);
}

char* char_array (long length) {
	return (char*) get_aligned_mem(length*sizeof(char), 16);
}


double* double_array (long length) {
	return (double*) get_aligned_mem(length*sizeof(double), 16);
}

float* float_array (long length) {
	return (float*) get_aligned_mem(length*sizeof(float), 16);
}

void** pointer_array (long length) {
	return (void**) get_aligned_mem(length*sizeof(void**), 16);
}

void free_array (void* ptr) {
	free_aligned_mem(ptr);
}

void free_pointers(void** ptrs, int n) {
	int i=0;
	for (i=0; i < n; i++) free_array(ptrs[n]);
}

void empty_printf(const char* s, ...) {
}

void int2double (double* dest, int* src, int length) {
	double* limit = dest + length;
	while(dest < limit) *dest++ = (double) *src++;
}

void char2double (double* dest, char* src, int length) {
	double* limit = dest + length;
	while(dest < limit) *dest++ = (double) *src++;
}

void uchar2double (double* dest, unsigned char* src, int length) {
	double* limit = dest + length;
	while(dest < limit) *dest++ = (double) *src++;
}

void double2int (int* dest, double* src, int length) {
	int* limit = dest + length;
	while (dest < limit) *dest++ = (int) *src++;
}

void long2double (double* dest, long* src, int length) {
	double* limit = dest + length;
	while (dest < limit) *dest++ = (double) *src++;
}

void double2long(long* dest, double* src, int length) {
	long* limit = dest + length;
	while(dest < limit) *dest++ = (long) *src++;
}

void double2char(char* dest, double* src, int length) {
	char* limit = dest + length;
	while(dest < limit) *dest++ = (char) *src++;
}

void double2uchar(unsigned char* dest, double* src, int length) {
	unsigned char* limit = dest + length;
	while(dest < limit) *dest++ = (unsigned char) *src++;
}


void double2float(float* dest, double* src, int length) {
	float* limit = dest + length;
	while(dest < limit) *dest++ = (float) *src++;
}

void float2double(double* dest, float* src, int length) {
	double* limit = dest + length;
	while(dest < limit) *dest++ = (double) *src++;
}



/*
	compare the sets A and B. Return in X the members of A that are not in B. Return in Y the members of B that are not in A.
	Returns vectors of indices I and J such that X=A(I) and Y=B(J)
	
*/
void set_diff (int* A, int* B, int A_length, int B_length, int* X, int* Y, int* X_length, int* Y_length, int* I, int* J) {

	int A_i = 0;
	int B_i = 0;

	int X_i = 0;
	int Y_i = 0;

	int cur_A = 0;
	int cur_B = 0;

	int* IA=int_array(A_length);
	int* IB=int_array(B_length);
	for(int i=0;i < A_length; i++) IA[i]=i;
	for(int i=0;i < B_length; i++) IB[i]=i;

	int* A_sort = int_array(A_length);
	memcpy(A_sort, A, A_length* sizeof(int));
	int* B_sort = int_array(B_length);
	memcpy(B_sort, B, B_length* sizeof(int));

	i_isort(A_sort, IA, A_length);
	i_isort(B_sort, IB, B_length);

	while ((A_i < A_length) || (B_i < B_length)) {

		cur_A = (A_i < A_length) ? A_sort[A_i] : A_sort[A_length - 1];
		cur_B = (B_i < B_length) ? B_sort[B_i] : B_sort[B_length - 1];

		if ((cur_A < cur_B) && (A_i < A_length)){

			X[X_i] = A_sort[A_i];
			if(I!=NULL) I[X_i]=IA[A_i];

			X_i++;
			A_i++;
		}
		else if ((cur_B < cur_A) && (B_i < B_length)){

			Y[Y_i] = B_sort[B_i];
			if(J!=NULL) J[Y_i]=IB[B_i];

			Y_i++;
			B_i++;
		}
		else {
			if ((A_i < A_length) && (B_i < B_length)) {
				A_i++;
				B_i++;
			}
			else if (A_i < A_length) {
				X[X_i] = A_sort[A_i];
				if(I) I[X_i]=IA[A_i];

				X_i++;
				A_i++;
			}
			else if (B_i < B_length) {
				Y[Y_i] = B_sort[B_i];
				if(J) J[Y_i]=IB[B_i];

				Y_i++;
				B_i++;
			}
		}
	}

	free_array(A_sort);
	free_array(B_sort);
	free_array(IA);
	free_array(IB);

	*X_length = X_i;
	*Y_length = Y_i;

	if(I) i_isort(I, NULL, *X_length);
	if(J) i_isort(J, NULL, *Y_length);

}

/*
	compare the sets A and B. Return in S the elements that are in A and B. Returns vectors of indices I and J such that A(I)=B(J) 
*/
void intersect (int* A, int* B, int A_length, int B_length, int* S, int* S_length, int* I, int*J) {

	int A_i = 0;
	int B_i = 0;

	int S_i = 0;
	int cur_A = 0;
	int cur_B = 0;

	int* IA = int_array(A_length);
	int* IB = int_array(B_length);
	for (int i=0; i < A_length; i++) IA[i] = i;
	for (int i=0; i < B_length; i++) IB[i] = i;

	int* A_sort = int_array(A_length);
	memcpy(A_sort, A, A_length*sizeof(int));
	int* B_sort = int_array(B_length);
	memcpy(B_sort, B, B_length*sizeof(int));

	i_isort(A_sort, IA, A_length);
	i_isort(B_sort, IB, B_length);

//	qsort((void*) A_sort, A_length, sizeof(int), &icompare);
//	qsort((void*) B_sort, B_length, sizeof(int), &icompare);

	while ((A_i < A_length) || (B_i < B_length)) {

		cur_A = (A_i < A_length) ? A_sort[A_i] : A_sort[A_length - 1];
		cur_B = (B_i < B_length) ? B_sort[B_i] : B_sort[B_length - 1];

		if ((cur_A < cur_B) && (A_i < A_length)){
			A_i++;
		}
		else if ((cur_B < cur_A) && (B_i < B_length)){
			B_i++;
		}
		else {
			if (cur_B == cur_A) {
				S[S_i] = cur_A;

				if ((I != NULL) && (J != NULL)) {
					I[S_i] = IA[A_i];
					J[S_i] = IB[B_i];
				}
				S_i++;
			}

			if (A_i < A_length) A_i++;
			if (B_i < B_length) B_i++;
		}
	}

	free_array(A_sort);
	free_array(B_sort);
	free_array(IA);
	free_array(IB);

	*S_length = S_i;
}

void bin_search (double val, double* x, int* left, int* right, int length) {
	if ((*left == -1) && (*right == -1)) {
		*left  =  0;
		*right = length-1;
	}

	if (*right - *left > 1) {
		int mid =  *left + (*right - *left)/2;

		if (x[mid] >= val) *right = mid;
		else *left = mid;

		bin_search(val, x, left, right, length);
	}
}

void linspace (double* dest, double pmin, double pmax, int count) {

	if (count < 2) {
		*dest = (pmin + pmax)/2.0;
		return;
	}

	int i=0;
	double step = 0;
	step = (pmax - pmin)/((double)(count-1));

	dest[0] = pmin;
	for (i=1; i < count; i++) dest[i] = dest[i-1] + step;
}

int is_little_endian (void) {
	int a = 1;

	char* p = (char*) &a;	

	return (*p == 1);
}	

void toogle_endianess (int* vals, int length) {
	int i=0;
	char* p = NULL;
	int tmp = 0;
	char* p2 = (char*) &tmp;

	for (i=0;i < length;i++) {
		p = (char*) (vals + i);	
		tmp = vals[i];	
		p[0] = p2[3];
		p[1] = p2[2];
		p[2] = p2[1]; 
		p[3] = p2[0];	
	}
}

/* find the equal error rate by binary search */
double equal_error_rate_rec (double* decision, double* cl, int length, double lbound, double ubound, double min_dist, 
			     double rIP_len, double rIN_len, double* r_b) {

/*	PRINTF("lbound: %g, ubound: %g\n", lbound, ubound); */

	double eer = -1;

	double* temp       = double_array(length);
	double* cur_labels = double_array(length);

	int* IN = int_array(length);
	int* IP = int_array(length); 

	// currently we are in the middle between lower- and upper bound
	double cur_threshold = lbound + (ubound - lbound)/2.0;

	// add current threshold to the decision values
	sadd(temp, decision, cur_threshold, length);

	// find those samples that are assigned to class -1 using the current threshold
	int IN_len = find_l   (IN, temp, 0, length);
	// find those samples that are assigned to class 1 using the current threshold
	int IP_len = find_geq (IP, temp, 0, length);

/*	PRINTF("IN_len: %i, IP_len: %i\n", IN_len, IP_len); */

	assign_i(cur_labels, IN, -1, IN_len); // current label assignment
	assign_i(cur_labels, IP,  1, IP_len);

	// temp is lower zero where a sample of class -1 was assigned to class 1
	// temp is greater zero where a sample of class 1 was assigned to class -1
	sub (temp, cl , cur_labels, length);

	// compute false positive rate
	double fpr = ((double)find_l(IN, temp, 0, length))/rIN_len; 
	// compute false negative rate
	double fnr = ((double)find_g(IP, temp, 0, length))/rIP_len;

	free_array(temp);
	free_array(IN);
	free_array(IP);
	free_array(cur_labels);

/*	PRINTF("abs(fnr - fpr): %g\n", fabs(fnr- fpr)); 
	PRINTF("1/length: %g\n", 1.0/((double) (rIN_len > rIP_len) ? rIN_len : rIP_len )); */
/*	PRINTF("(ubound - lbound): %g\n", ubound - lbound); 
	PRINTF("lbound: %g, cur_threshold: %g\n", lbound, cur_threshold);
	PRINTF("min_dist/2: %g\n", min_dist/2);*/

	if (((ubound - lbound)/2.0 <= min_dist) || (fabs(fpr-fnr) < 1.0/((double) (rIN_len > rIP_len) ? rIN_len : rIP_len))){ /* if maximal precision with respect to the decision values was reached ... quit */
		*r_b = cur_threshold;
		eer = (fpr + fnr) / 2;
	}
	else {
		if (fpr > fnr) { // to many samples were assigned to class 1 => lower the threshold
			ubound = cur_threshold;
			eer = equal_error_rate_rec (decision, cl, length, lbound, ubound,  min_dist, rIP_len, rIN_len, r_b);
		}
		else if (fpr < fnr) { // to many samples were assigned to class -1 => increase the threshold
			lbound = cur_threshold;
			eer = equal_error_rate_rec (decision, cl, length, lbound , ubound, min_dist, rIP_len, rIN_len, r_b);
		}
		else {
			*r_b = cur_threshold;
			eer = (fpr + fnr) / 2; // we finished ...
		}
	}

	return eer;
}

double equal_error_rate (double* decision, double* cl, int length, double plabel, double nlabel, double* r_b) {
	int i=0;

/*	print_array("decision: ", decision, length); */

	// create copy of decision values
	double* temp = double_array(length);
	memcpy(temp, decision, length*sizeof(double));
	// sort copy of decision values
	qsort((void*) temp, length, sizeof(double), &dcompare);

	// determine interval of the decision values
	double ubound = (fabs(temp[length-1])  > fabs(temp[0])) ? fabs(temp[length-1]) : fabs(temp[0]);
	double lbound = -ubound; 

/*	PRINTF ("lbound start: %g, ubound start: %g\n", lbound, ubound); */

	// determine the smallest distance between two different decision values
	double min_dist = ubound - lbound;
	for (i=0; i < length - 1; i++) {
		double dist = fabs(temp[i] - temp[i+1]);

		min_dist = ((dist < min_dist) && (dist > 0))? dist : min_dist;
	}

/*	PRINTF("min_dist: %g\n", min_dist); */

	int* IP = int_array(length);
	int* IN = int_array(length);
	
	// find decision values to consider
	int IP_len = find_eq(IP, cl, plabel, length);
	int IN_len = find_eq(IN, cl, nlabel, length);

/*	PRINTF("IP_len: %i, IN_len: %i\n", IP_len, IN_len); */
	
	double* tdecision = double_array(IP_len + IN_len);
	double* tcl 	  = double_array(IP_len + IN_len);

	// get these decision values
	get_i(tdecision         , decision, IP, IP_len);
	get_i(tdecision + IP_len, decision, IN, IN_len);

	// create temporary class labels in {1,-1}
	for(i = 0; i < IP_len; i++)              tcl[i] = 1;
	for(i = IP_len; i < IP_len + IN_len;i++) tcl[i] = -1;

/*	print_array("tdecision: ", tdecision, IP_len + IN_len);
	print_array("tcl: ", tcl, IP_len + IN_len); */

	// compute equal error rate
	double eer = equal_error_rate_rec (tdecision, tcl, IP_len + IN_len, lbound, ubound, min_dist, IP_len, IN_len, r_b);

	free_array(IP);
	free_array(IN);
	free_array(tdecision);
	free_array(tcl);
	free_array(temp);

	return eer;
}
